# WARNING: this file is auto-generated by 'async_to_sync.py'
# from the original file 'test_copy_async.py'
# DO NOT CHANGE! Change the original file instead.
import string
import hashlib
from io import BytesIO, StringIO
from random import choice, randrange
from itertools import cycle

import pytest

import psycopg
from psycopg import errors as e
from psycopg import pq, sql
from psycopg.abc import Buffer
from psycopg.copy import Copy, LibpqWriter, QueuedLibpqWriter
from psycopg.adapt import Dumper, PyFormat
from psycopg.types import TypeInfo
from psycopg.types.hstore import register_hstore
from psycopg.types.numeric import Int4

from .utils import eur
from .acompat import Event, gather, spawn
from ._test_copy import sample_binary  # noqa: F401
from ._test_copy import FileWriter, ensure_table, py_to_raw, sample_binary_rows
from ._test_copy import sample_records, sample_tabledef, sample_text, sample_values
from ._test_copy import special_chars
from .test_adapt import StrNoneBinaryDumper, StrNoneDumper

pytestmark = pytest.mark.crdb_skip("copy")


@pytest.mark.parametrize("format", pq.Format)
def test_copy_out_read(conn, format):
    if format == pq.Format.TEXT:
        want = [row + b"\n" for row in sample_text.splitlines()]
    else:
        want = sample_binary_rows

    cur = conn.cursor()
    with cur.copy(f"copy ({sample_values}) to stdout (format {format.name})") as copy:
        for row in want:
            got = copy.read()
            assert got == row
            assert conn.info.transaction_status == pq.TransactionStatus.ACTIVE

        assert copy.read() == b""
        assert copy.read() == b""

    assert copy.read() == b""
    assert conn.info.transaction_status == pq.TransactionStatus.INTRANS


@pytest.mark.parametrize("format", pq.Format)
@pytest.mark.parametrize("row_factory", ["tuple_row", "dict_row", "namedtuple_row"])
def test_copy_out_iter(conn, format, row_factory):
    if format == pq.Format.TEXT:
        want = [row + b"\n" for row in sample_text.splitlines()]
    else:
        want = sample_binary_rows

    rf = getattr(psycopg.rows, row_factory)
    cur = conn.cursor(row_factory=rf)
    with cur.copy(f"copy ({sample_values}) to stdout (format {format.name})") as copy:
        assert list(copy) == want

    assert conn.info.transaction_status == pq.TransactionStatus.INTRANS


@pytest.mark.parametrize("format", pq.Format)
@pytest.mark.parametrize("row_factory", ["tuple_row", "dict_row", "namedtuple_row"])
def test_copy_out_no_result(conn, format, row_factory):
    rf = getattr(psycopg.rows, row_factory)
    cur = conn.cursor(row_factory=rf)
    with cur.copy(f"copy ({sample_values}) to stdout (format {format.name})"):
        with pytest.raises(e.ProgrammingError):
            cur.fetchone()


@pytest.mark.parametrize("ph, params", [("%s", (10,)), ("%(n)s", {"n": 10})])
def test_copy_out_param(conn, ph, params):
    cur = conn.cursor()
    with cur.copy(
        f"copy (select * from generate_series(1, {ph})) to stdout", params
    ) as copy:
        copy.set_types(["int4"])
        assert list(copy.rows()) == [(i + 1,) for i in range(10)]

    assert conn.info.transaction_status == pq.TransactionStatus.INTRANS


@pytest.mark.parametrize("format", pq.Format)
@pytest.mark.parametrize("typetype", ["names", "oids"])
def test_read_rows(conn, format, typetype):
    cur = conn.cursor()
    with cur.copy("""
        copy (
            select 10::int4, 'hello'::text, '{0.0,1.0}'::float8[]
    ) to stdout (format %s)""" % format.name) as copy:
        copy.set_types(["int4", "text", "float8[]"])
        row = copy.read_row()
        assert copy.read_row() is None

    assert row == (10, "hello", [0.0, 1.0])
    assert conn.info.transaction_status == pq.TransactionStatus.INTRANS


@pytest.mark.parametrize("format", pq.Format)
def test_rows(conn, format):
    cur = conn.cursor()
    with cur.copy(f"copy ({sample_values}) to stdout (format {format.name})") as copy:
        copy.set_types(["int4", "int4", "text"])
        rows = list(copy.rows())

    assert rows == sample_records
    assert conn.info.transaction_status == pq.TransactionStatus.INTRANS


@pytest.mark.parametrize("format", pq.Format)
def test_set_types(conn, format):
    sample = ({"foo": "bar"}, 123)
    cur = conn.cursor()
    ensure_table(cur, "id serial primary key, data jsonb, data2 bigint")
    with cur.copy(
        f"copy copy_in (data, data2) from stdin (format {format.name})"
    ) as copy:
        copy.set_types(["jsonb", "bigint"])
        copy.write_row(sample)
    cur.execute("select data, data2 from copy_in")
    data = cur.fetchone()
    assert data == sample


@pytest.mark.parametrize("format", pq.Format)
@pytest.mark.parametrize("use_set_types", [True, False])
def test_rowlen_mismatch(conn, format, use_set_types):
    samples = [["foo", "bar"], ["foo", "bar", "baz"]]
    cur = conn.cursor()
    ensure_table(cur, "id serial primary key, data text, data2 text")
    with pytest.raises(psycopg.DataError):
        with cur.copy(
            f"copy copy_in (data, data2) from stdin (format {format.name})"
        ) as copy:
            if use_set_types:
                copy.set_types(["text", "text"])
            for row in samples:
                copy.write_row(row)


def test_set_custom_type(conn, hstore):
    command = """copy (select '"a"=>"1", "b"=>"2"'::hstore) to stdout"""
    cur = conn.cursor()

    with cur.copy(command) as copy:
        rows = list(copy.rows())

    assert rows == [('"a"=>"1", "b"=>"2"',)]

    register_hstore(TypeInfo.fetch(conn, "hstore"), cur)
    with cur.copy(command) as copy:
        copy.set_types(["hstore"])
        rows = list(copy.rows())

    assert rows == [({"a": "1", "b": "2"},)]


@pytest.mark.parametrize("format", pq.Format)
def test_copy_out_allchars(conn, format):
    cur = conn.cursor()
    chars = list(map(chr, range(1, 256))) + [eur]
    conn.execute("set client_encoding to utf8")
    rows = []
    query = sql.SQL("copy (select unnest({}::text[])) to stdout (format {})").format(
        chars, sql.SQL(format.name)
    )
    with cur.copy(query) as copy:
        copy.set_types(["text"])
        while row := copy.read_row():
            assert len(row) == 1
            rows.append(row[0])

    assert rows == chars


@pytest.mark.parametrize("format", pq.Format)
def test_read_row_notypes(conn, format):
    cur = conn.cursor()
    with cur.copy(f"copy ({sample_values}) to stdout (format {format.name})") as copy:
        rows = []
        while row := copy.read_row():
            rows.append(row)

    ref = [tuple((py_to_raw(i, format) for i in record)) for record in sample_records]
    assert rows == ref


@pytest.mark.parametrize("format", pq.Format)
def test_rows_notypes(conn, format):
    cur = conn.cursor()
    with cur.copy(f"copy ({sample_values}) to stdout (format {format.name})") as copy:
        rows = list(copy.rows())
    ref = [tuple((py_to_raw(i, format) for i in record)) for record in sample_records]
    assert rows == ref


@pytest.mark.parametrize("err", [-1, 1])
@pytest.mark.parametrize("format", pq.Format)
def test_copy_out_badntypes(conn, format, err):
    cur = conn.cursor()
    with cur.copy(f"copy ({sample_values}) to stdout (format {format.name})") as copy:
        copy.set_types([0] * (len(sample_records[0]) + err))
        with pytest.raises(e.ProgrammingError):
            copy.read_row()


@pytest.mark.parametrize(
    "format, buffer",
    [(pq.Format.TEXT, "sample_text"), (pq.Format.BINARY, "sample_binary")],
)
def test_copy_in_buffers(conn, format, buffer):
    cur = conn.cursor()
    ensure_table(cur, sample_tabledef)
    with cur.copy(f"copy copy_in from stdin (format {format.name})") as copy:
        copy.write(globals()[buffer])

    cur.execute("select * from copy_in order by 1")
    data = cur.fetchall()
    assert data == sample_records


def test_copy_in_buffers_pg_error(conn):
    cur = conn.cursor()
    ensure_table(cur, sample_tabledef)
    with pytest.raises(e.UniqueViolation):
        with cur.copy("copy copy_in from stdin (format text)") as copy:
            copy.write(sample_text)
            copy.write(sample_text)
    assert conn.info.transaction_status == pq.TransactionStatus.INERROR


def test_copy_bad_result(conn):
    conn.set_autocommit(True)

    cur = conn.cursor()

    with pytest.raises(e.SyntaxError):
        with cur.copy("wat"):
            pass

    with pytest.raises(e.ProgrammingError):
        with cur.copy("select 1"):
            pass

    with pytest.raises(e.ProgrammingError):
        with cur.copy("reset timezone"):
            pass

    with pytest.raises(e.ProgrammingError):
        with cur.copy("copy (select 1) to stdout; select 1") as copy:
            list(copy)

    with pytest.raises(e.ProgrammingError):
        with cur.copy("select 1; copy (select 1) to stdout"):
            pass


def test_copy_in_str(conn):
    cur = conn.cursor()
    ensure_table(cur, sample_tabledef)
    with cur.copy("copy copy_in from stdin (format text)") as copy:
        copy.write(sample_text.decode())

    cur.execute("select * from copy_in order by 1")
    data = cur.fetchall()
    assert data == sample_records


def test_copy_in_error(conn):
    cur = conn.cursor()
    ensure_table(cur, sample_tabledef)
    with pytest.raises(TypeError):
        with cur.copy("copy copy_in from stdin (format binary)") as copy:
            copy.write(sample_text.decode())

    assert conn.info.transaction_status == pq.TransactionStatus.INERROR


@pytest.mark.parametrize("format", pq.Format)
def test_copy_in_empty(conn, format):
    cur = conn.cursor()
    ensure_table(cur, sample_tabledef)
    with cur.copy(f"copy copy_in from stdin (format {format.name})"):
        pass

    assert conn.info.transaction_status == pq.TransactionStatus.INTRANS
    assert cur.rowcount == 0


@pytest.mark.slow
def test_copy_big_size_record(conn):
    cur = conn.cursor()
    ensure_table(cur, sample_tabledef)
    data = "".join((chr(randrange(1, 256)) for i in range(10 * 1024 * 1024)))
    with cur.copy("copy copy_in (data) from stdin") as copy:
        copy.write_row([data])

    cur.execute("select data from copy_in limit 1")
    assert cur.fetchone() == (data,)


@pytest.mark.slow
@pytest.mark.parametrize("pytype", [str, bytes, bytearray, memoryview])
def test_copy_big_size_block(conn, pytype):
    cur = conn.cursor()
    ensure_table(cur, sample_tabledef)
    data = "".join((choice(string.ascii_letters) for i in range(10 * 1024 * 1024)))
    copy_data = data + "\n" if pytype is str else pytype(data.encode() + b"\n")
    with cur.copy("copy copy_in (data) from stdin") as copy:
        copy.write(copy_data)

    cur.execute("select data from copy_in limit 1")
    assert cur.fetchone() == (data,)


@pytest.mark.parametrize("format", pq.Format)
def test_subclass_adapter(conn, format):
    if format == pq.Format.TEXT:
        from psycopg.types.string import StrDumper as BaseDumper
    else:
        from psycopg.types.string import StrBinaryDumper

        BaseDumper = StrBinaryDumper  # type: ignore

    class MyStrDumper(BaseDumper):

        def dump(self, obj):
            rv = super().dump(obj)
            assert rv
            return bytes(rv) * 2

    conn.adapters.register_dumper(str, MyStrDumper)

    cur = conn.cursor()
    ensure_table(cur, sample_tabledef)

    with cur.copy(f"copy copy_in (data) from stdin (format {format.name})") as copy:
        copy.write_row(("hello",))

    cur.execute("select data from copy_in")
    rec = cur.fetchone()
    assert rec[0] == "hellohello"


@pytest.mark.parametrize("format", pq.Format)
def test_subclass_nulling_dumper(conn, format):
    Base: type = StrNoneDumper if format == pq.Format.TEXT else StrNoneBinaryDumper

    class MyStrDumper(Base):  # type: ignore

        def dump(self, obj):
            return super().dump(obj) if obj else None

    conn.adapters.register_dumper(str, MyStrDumper)

    cur = conn.cursor()
    ensure_table(cur, sample_tabledef)

    with cur.copy(f"copy copy_in (data) from stdin (format {format.name})") as copy:
        copy.write_row(("hello",))
        copy.write_row(("",))

    cur.execute("select data from copy_in order by col1")
    recs = cur.fetchall()
    assert recs == [("hello",), (None,)]


@pytest.mark.parametrize("format", pq.Format)
def test_copy_in_error_empty(conn, format):
    cur = conn.cursor()
    ensure_table(cur, sample_tabledef)
    with pytest.raises(ZeroDivisionError, match="mannaggiamiseria"):
        with cur.copy(f"copy copy_in from stdin (format {format.name})"):
            raise ZeroDivisionError("mannaggiamiseria")

    assert conn.info.transaction_status == pq.TransactionStatus.INERROR


def test_copy_in_buffers_with_pg_error(conn):
    cur = conn.cursor()
    ensure_table(cur, sample_tabledef)
    with pytest.raises(e.UniqueViolation):
        with cur.copy("copy copy_in from stdin (format text)") as copy:
            copy.write(sample_text)
            copy.write(sample_text)

    assert conn.info.transaction_status == pq.TransactionStatus.INERROR


def test_copy_in_buffers_with_py_error(conn):
    cur = conn.cursor()
    ensure_table(cur, sample_tabledef)
    with pytest.raises(ZeroDivisionError, match="nuttengoggenio"):
        with cur.copy("copy copy_in from stdin (format text)") as copy:
            copy.write(sample_text)
            raise ZeroDivisionError("nuttengoggenio")

    assert conn.info.transaction_status == pq.TransactionStatus.INERROR


def test_copy_out_error_with_copy_finished(conn):
    cur = conn.cursor()
    with pytest.raises(ZeroDivisionError):
        with cur.copy("copy (select generate_series(1, 2)) to stdout") as copy:
            copy.read_row()
            1 / 0

    assert conn.info.transaction_status == pq.TransactionStatus.INTRANS


def test_copy_out_error_with_copy_not_finished(conn):
    cur = conn.cursor()
    with pytest.raises(ZeroDivisionError):
        with cur.copy("copy (select generate_series(1, 1000000)) to stdout") as copy:
            copy.read_row()
            1 / 0

    assert conn.info.transaction_status == pq.TransactionStatus.INERROR


def test_copy_out_server_error(conn):
    cur = conn.cursor()
    with pytest.raises(e.DivisionByZero):
        with cur.copy(
            "copy (select 1/n from generate_series(-10, 10) x(n)) to stdout"
        ) as copy:
            for block in copy:
                pass

    assert conn.info.transaction_status == pq.TransactionStatus.INERROR


@pytest.mark.parametrize("format", pq.Format)
def test_copy_in_records(conn, format):
    cur = conn.cursor()
    ensure_table(cur, sample_tabledef)

    with cur.copy(f"copy copy_in from stdin (format {format.name})") as copy:
        for row in sample_records:
            if format == pq.Format.BINARY:
                row2 = tuple((Int4(i) if isinstance(i, int) else i for i in row))
                row = row2  # type: ignore[assignment]
            copy.write_row(row)

    cur.execute("select * from copy_in order by 1")
    data = cur.fetchall()
    assert data == sample_records


@pytest.mark.parametrize("format", pq.Format)
def test_copy_in_records_set_types(conn, format):
    cur = conn.cursor()
    ensure_table(cur, sample_tabledef)

    with cur.copy(f"copy copy_in from stdin (format {format.name})") as copy:
        copy.set_types(["int4", "int4", "text"])
        for row in sample_records:
            copy.write_row(row)

    cur.execute("select * from copy_in order by 1")
    data = cur.fetchall()
    assert data == sample_records


@pytest.mark.parametrize("format", pq.Format)
def test_copy_in_records_binary(conn, format):
    cur = conn.cursor()
    ensure_table(cur, "col1 serial primary key, col2 int, data text")

    with cur.copy(
        f"copy copy_in (col2, data) from stdin (format {format.name})"
    ) as copy:
        for row in sample_records:
            copy.write_row((None, row[2]))

    cur.execute("select * from copy_in order by 1")
    data = cur.fetchall()
    assert data == [(1, None, "hello"), (2, None, "world")]


class StrictIntDumper(Dumper):
    oid = psycopg.adapters.types["int4"].oid

    def dump(self, obj: int) -> Buffer:
        if type(obj) is not int:
            raise TypeError(f"bad type: {obj!r}")
        return str(obj).encode()


def test_copy_in_text_no_pinning(conn):
    cur = conn.cursor()
    cur.adapters.register_dumper(int, StrictIntDumper)

    cols = [
        "col1 serial primary key",
        "col2 int",
        "col3 int",
        "col4 double precision",
        "col5 double precision",
    ]
    ensure_table(cur, ",".join(cols))

    with cur.copy(
        "copy copy_in (col2,col3,col4,col5) from stdin (format text)"
    ) as copy:
        # no pinned dumpers: type check & cast done on postgres side
        # allows to mix castable reprs more freely
        # slower than pinned, late errors from postgres jeopardizing copy cursor
        copy.write_row([1, "2", 3, "4.1"])
        copy.write_row(["1", 2, 3.0, 4])

    cur.execute("select col2,col3,col4,col5 from copy_in order by 1")
    data = cur.fetchall()
    assert data == [(1, 2, 3, 4.1), (1, 2, 3, 4)]


def test_copy_in_text_pinned(conn):
    cur = conn.cursor()
    cur.adapters.register_dumper(int, StrictIntDumper)

    cols = [
        "col1 serial primary key",
        "col2 int",
        "col3 int",
        "col4 double precision",
        "col5 double precision",
    ]
    ensure_table(cur, ",".join(cols))

    with cur.copy(
        "copy copy_in (col2,col3,col4,col5) from stdin (format text)"
    ) as copy:
        # pinned dumpers from set_types: type check & cast done on psycopg side
        # much faster, allows catching errors early without postgres involvement
        copy.set_types(["int4", "int4", "double precision", "double precision"])
        copy.write_row([1, 2, 3, 4.1])
        with pytest.raises(
            (e.DataError, TypeError)
        ):  # FIXME: should errors from dumpers be harmonized?
            copy.write_row([1.0, 2, 3, 4.1])
        with pytest.raises((e.DataError, TypeError)):
            copy.write_row([1, "2", 3, 4.1])

    cur.execute("select col2,col3,col4,col5 from copy_in order by 1")
    data = cur.fetchall()
    assert data == [(1, 2, 3, 4.1)]


def test_copy_in_allchars(conn):
    cur = conn.cursor()
    ensure_table(cur, sample_tabledef)

    conn.execute("set client_encoding to utf8")
    with cur.copy("copy copy_in from stdin (format text)") as copy:
        for i in range(1, 256):
            copy.write_row((i, None, chr(i)))
        copy.write_row((ord(eur), None, eur))

    cur.execute("""
select col1 = ascii(data), col2 is null, length(data), count(*)
from copy_in group by 1, 2, 3
""")
    data = cur.fetchall()
    assert data == [(True, True, 1, 256)]


def test_copy_in_format(conn):
    file = BytesIO()
    conn.execute("set client_encoding to utf8")
    cur = conn.cursor()
    with Copy(cur, writer=FileWriter(file)) as copy:
        for i in range(1, 256):
            copy.write_row((i, chr(i)))

    file.seek(0)
    rows = file.read().split(b"\n")
    assert not rows[-1]
    del rows[-1]

    for i, row in enumerate(rows, start=1):
        fields = row.split(b"\t")
        assert len(fields) == 2
        assert int(fields[0].decode()) == i
        if i in special_chars:
            assert fields[1].decode() == f"\\{special_chars[i]}"
        else:
            assert fields[1].decode() == chr(i)


@pytest.mark.parametrize(
    "format, buffer",
    [(pq.Format.TEXT, "sample_text"), (pq.Format.BINARY, "sample_binary")],
)
def test_file_writer(conn, format, buffer):
    file = BytesIO()
    conn.execute("set client_encoding to utf8")
    cur = conn.cursor()
    with Copy(cur, binary=format, writer=FileWriter(file)) as copy:
        for record in sample_records:
            copy.write_row(record)

    file.seek(0)
    want = globals()[buffer]
    got = file.read()
    assert got == want


@pytest.mark.slow
def test_copy_from_to(conn):
    # Roundtrip from file to database to file blockwise
    gen = DataGenerator(conn, nrecs=1024, srec=10 * 1024)
    gen.ensure_table()
    cur = conn.cursor()
    with cur.copy("copy copy_in from stdin") as copy:
        for block in gen.blocks():
            copy.write(block)

    gen.assert_data()

    f = BytesIO()
    with cur.copy("copy copy_in to stdout") as copy:
        for block in copy:
            f.write(block)

    f.seek(0)
    assert gen.sha(f) == gen.sha(gen.file())


@pytest.mark.slow
@pytest.mark.parametrize("pytype", [bytes, bytearray, memoryview])
def test_copy_from_to_bytes(conn, pytype):
    # Roundtrip from file to database to file blockwise
    gen = DataGenerator(conn, nrecs=1024, srec=10 * 1024)
    gen.ensure_table()
    cur = conn.cursor()
    with cur.copy("copy copy_in from stdin") as copy:
        for block in gen.blocks():
            copy.write(pytype(block.encode()))

    gen.assert_data()

    f = BytesIO()
    with cur.copy("copy copy_in to stdout") as copy:
        for block in copy:
            f.write(block)

    f.seek(0)
    assert gen.sha(f) == gen.sha(gen.file())


@pytest.mark.slow
def test_copy_from_insane_size(conn):
    # Trying to trigger a "would block" error
    gen = DataGenerator(
        conn, nrecs=4 * 1024, srec=10 * 1024, block_size=20 * 1024 * 1024
    )
    gen.ensure_table()
    cur = conn.cursor()
    with cur.copy("copy copy_in from stdin") as copy:
        for block in gen.blocks():
            copy.write(block)

    gen.assert_data()


def test_copy_rowcount(conn):
    gen = DataGenerator(conn, nrecs=3, srec=10)
    gen.ensure_table()

    cur = conn.cursor()
    with cur.copy("copy copy_in from stdin") as copy:
        for block in gen.blocks():
            copy.write(block)
    assert cur.rowcount == 3

    gen = DataGenerator(conn, nrecs=2, srec=10, offset=3)
    with cur.copy("copy copy_in from stdin") as copy:
        for rec in gen.records():
            copy.write_row(rec)
    assert cur.rowcount == 2

    with cur.copy("copy copy_in to stdout") as copy:
        for block in copy:
            pass
    assert cur.rowcount == 5

    with pytest.raises(e.BadCopyFileFormat):
        with cur.copy("copy copy_in (id) from stdin") as copy:
            for rec in gen.records():
                copy.write_row(rec)
    assert cur.rowcount == -1


def test_copy_query(conn):
    cur = conn.cursor()
    with cur.copy("copy (select 1) to stdout") as copy:
        assert cur._query.query == b"copy (select 1) to stdout"
        assert not cur._query.params
        list(copy)


def test_cant_reenter(conn):
    cur = conn.cursor()
    with cur.copy("copy (select 1) to stdout") as copy:
        list(copy)

    with pytest.raises(TypeError):
        with copy:
            list(copy)


def test_str(conn):
    cur = conn.cursor()
    with cur.copy("copy (select 1) to stdout") as copy:
        assert "[ACTIVE]" in str(copy)
        list(copy)

    assert "[INTRANS]" in str(copy)


def test_description(conn):
    with conn.cursor() as cur:
        with cur.copy("copy (select 'This', 'Is', 'Text') to stdout") as copy:
            len(cur.description) == 3
            assert cur.description[0].name == "column_1"
            assert cur.description[2].name == "column_3"
            list(copy.rows())

        len(cur.description) == 3
        assert cur.description[0].name == "column_1"
        assert cur.description[2].name == "column_3"


def test_binary_partial_row(conn):
    cur = conn.cursor()
    ensure_table(cur, "id serial primary key, num int4, arr int4[][]")
    with pytest.raises(
        psycopg.DataError, match="nested lists have inconsistent depths"
    ):
        with cur.copy("copy copy_in (num, arr) from stdin (format binary)") as copy:
            copy.set_types(["int4", "int4[]"])
            copy.write_row([15, None])
            copy.write_row([16, [[None], None]])


@pytest.mark.parametrize("format", pq.Format)
def test_clean_buffer_on_error(conn, format):
    cur = conn.cursor()
    ensure_table(cur, "id serial primary key, num int4, obj jsonb")
    with cur.copy(f"copy copy_in (num, obj) from stdin (format {format.name})") as copy:
        copy.set_types(["int4", "jsonb"])
        copy.write_row([15, {}])
        with pytest.raises(TypeError):
            copy.write_row([16, 1j])
        copy.write_row([17, []])

    cur.execute("select num, obj from copy_in order by id")
    assert cur.fetchall() == [(15, {}), (17, [])]


@pytest.mark.parametrize(
    "format, buffer",
    [(pq.Format.TEXT, "sample_text"), (pq.Format.BINARY, "sample_binary")],
)
def test_worker_life(conn, format, buffer):
    cur = conn.cursor()
    ensure_table(cur, sample_tabledef)
    with cur.copy(
        f"copy copy_in from stdin (format {format.name})", writer=QueuedLibpqWriter(cur)
    ) as copy:
        assert not copy.writer._worker
        copy.write(globals()[buffer])
        assert copy.writer._worker

    assert not copy.writer._worker
    cur.execute("select * from copy_in order by 1")
    data = cur.fetchall()
    assert data == sample_records


def test_worker_error_propagated(conn, monkeypatch):

    def copy_to_broken(pgconn, buffer, flush=True):
        raise ZeroDivisionError
        yield

    monkeypatch.setattr(psycopg._copy, "copy_to", copy_to_broken)
    cur = conn.cursor()
    cur.execute("create temp table wat (a text, b text)")
    with pytest.raises(ZeroDivisionError):
        with cur.copy("copy wat from stdin", writer=QueuedLibpqWriter(cur)) as copy:
            copy.write("a,b")


@pytest.mark.parametrize(
    "format, buffer",
    [(pq.Format.TEXT, "sample_text"), (pq.Format.BINARY, "sample_binary")],
)
def test_connection_writer(conn, format, buffer):
    cur = conn.cursor()
    writer = LibpqWriter(cur)

    ensure_table(cur, sample_tabledef)
    with cur.copy(
        f"copy copy_in from stdin (format {format.name})", writer=writer
    ) as copy:
        assert copy.writer is writer
        copy.write(globals()[buffer])

    cur.execute("select * from copy_in order by 1")
    data = cur.fetchall()
    assert data == sample_records


@pytest.mark.slow
@pytest.mark.parametrize(
    "fmt, set_types",
    [(pq.Format.TEXT, True), (pq.Format.TEXT, False), (pq.Format.BINARY, True)],
)
@pytest.mark.parametrize("method", ["read", "iter", "row", "rows"])
def test_copy_to_leaks(conn_cls, dsn, faker, fmt, set_types, method, gc):
    faker.format = PyFormat.from_pq(fmt)
    faker.choose_schema(ncols=20)
    faker.make_records(20)

    def work():
        with conn_cls.connect(dsn) as conn:
            with conn.cursor(binary=fmt) as cur:
                cur.execute(faker.drop_stmt)
                cur.execute(faker.create_stmt)
                with faker.find_insert_problem(conn):
                    cur.executemany(faker.insert_stmt, faker.records)

                stmt = sql.SQL(
                    "copy (select {} from {} order by id) to stdout (format {})"
                ).format(
                    sql.SQL(", ").join(faker.fields_names),
                    faker.table_name,
                    sql.SQL(fmt.name),
                )

                with cur.copy(stmt) as copy:
                    if set_types:
                        copy.set_types(faker.types_names)

                    if method == "read":
                        while copy.read():
                            pass
                    elif method == "iter":
                        list(copy)
                    elif method == "row":
                        while copy.read_row() is not None:
                            pass
                    elif method == "rows":
                        list(copy.rows())

    gc.collect()
    n = []
    for i in range(3):
        work()
        gc.collect()
        n.append(gc.count())

    assert n[0] == n[1] == n[2], f"objects leaked: {n[1] - n[0]}, {n[2] - n[1]}"


@pytest.mark.slow
@pytest.mark.parametrize(
    "fmt, set_types",
    [(pq.Format.TEXT, True), (pq.Format.TEXT, False), (pq.Format.BINARY, True)],
)
def test_copy_from_leaks(conn_cls, dsn, faker, fmt, set_types, gc):
    faker.format = PyFormat.from_pq(fmt)
    faker.choose_schema(ncols=20)
    faker.make_records(20)

    def work():
        with conn_cls.connect(dsn) as conn:
            with conn.cursor(binary=fmt) as cur:
                cur.execute(faker.drop_stmt)
                cur.execute(faker.create_stmt)

                stmt = sql.SQL("copy {} ({}) from stdin (format {})").format(
                    faker.table_name,
                    sql.SQL(", ").join(faker.fields_names),
                    sql.SQL(fmt.name),
                )
                with cur.copy(stmt) as copy:
                    if set_types:
                        copy.set_types(faker.types_names)
                    for row in faker.records:
                        copy.write_row(row)

                cur.execute(faker.select_stmt)
                recs = cur.fetchall()

                for got, want in zip(recs, faker.records):
                    faker.assert_record(got, want)

    gc.collect()
    n = []
    for i in range(3):
        work()
        gc.collect()
        n.append(gc.count())

    assert n[0] == n[1] == n[2], f"objects leaked: {n[1] - n[0]}, {n[2] - n[1]}"


@pytest.mark.slow
@pytest.mark.parametrize("mode", ["row", "block", "binary"])
def test_copy_table_across(conn_cls, dsn, faker, mode):
    faker.choose_schema(ncols=20)
    faker.make_records(20)

    connect = conn_cls.connect
    with connect(dsn) as conn1, connect(dsn) as conn2:
        faker.table_name = sql.Identifier("copy_src")
        conn1.execute(faker.drop_stmt)
        conn1.execute(faker.create_stmt)
        conn1.cursor().executemany(faker.insert_stmt, faker.records)

        faker.table_name = sql.Identifier("copy_tgt")
        conn2.execute(faker.drop_stmt)
        conn2.execute(faker.create_stmt)

        fmt = "(format binary)" if mode == "binary" else ""
        with conn1.cursor().copy(f"copy copy_src to stdout {fmt}") as copy1:
            with conn2.cursor().copy(f"copy copy_tgt from stdin {fmt}") as copy2:
                if mode == "row":
                    for row in copy1.rows():
                        copy2.write_row(row)
                else:
                    for data in copy1:
                        copy2.write(data)

        cur = conn2.execute(faker.select_stmt)
        recs = cur.fetchall()
        for got, want in zip(recs, faker.records):
            faker.assert_record(got, want)


def test_copy_concurrency(conn):
    """
    Test that copy operations hold the connection lock for the entire operation.

    This test verifies the fix for the concurrency issue where Cursor.copy()
    was not holding the connection lock throughout the copy context, allowing
    concurrent operations to interfere.
    """
    conn.execute("create temp table copy_concurrency_test (id int, data text)")

    # Events to coordinate execution between copy task and workers
    copy_entered = Event()
    wrote_first = Event()
    wrote_second = Event()
    can_proceed = Event()

    # Track execution order to verify workers run after copy completes
    execution_log = []

    def copy_task():
        """Copy task that writes two rows with controlled pauses."""
        cur = conn.cursor()
        with cur.copy("copy copy_concurrency_test from stdin") as copy:
            # Pause after entering copy context
            execution_log.append("entered_copy")
            copy_entered.set()
            can_proceed.wait()

            # Write first row and pause
            copy.write_row((1, "first"))
            execution_log.append("wrote_row_1")
            wrote_first.set()
            can_proceed.wait()

            # Write second row and pause
            copy.write_row((2, "second"))
            execution_log.append("wrote_row_2")
            wrote_second.set()
            can_proceed.wait()

        # Copy context exited, lock should now be released
        execution_log.append("exited_copy")

    def worker_task():
        """
        Worker that attempts to execute a query on a different cursor.
        Should block until copy completes due to connection lock.
        """
        # Try to execute on another cursor - this should block until copy exits
        worker_cur = conn.cursor()
        worker_cur.execute("select 1")
        execution_log.append("worker_completed")

    # Start the copy task
    t_copy = spawn(copy_task)

    # Wait for copy to enter, then spawn first worker
    copy_entered.wait()
    t_worker1 = spawn(worker_task)

    # Allow copy to proceed to write first row
    can_proceed.set()
    can_proceed.clear()
    wrote_first.wait()

    # Spawn second worker after first row
    t_worker2 = spawn(worker_task)

    # Allow copy to proceed to write second row
    can_proceed.set()
    can_proceed.clear()
    wrote_second.wait()

    # Spawn third worker after second row
    t_worker3 = spawn(worker_task)

    # Allow copy to exit
    can_proceed.set()

    # Wait for all tasks to complete
    gather(t_copy, t_worker1, t_worker2, t_worker3)

    # Verify the data was written correctly
    cur = conn.execute("select * from copy_concurrency_test order by id")
    rows = cur.fetchall()
    assert rows == [(1, "first"), (2, "second")]

    # Verify that all workers completed AFTER copy exited
    assert execution_log == [
        "entered_copy",
        "wrote_row_1",
        "wrote_row_2",
        "exited_copy",
        "worker_completed",
        "worker_completed",
        "worker_completed",
    ]


class DataGenerator:

    def __init__(self, conn, nrecs, srec, offset=0, block_size=8192):
        self.conn = conn
        self.nrecs = nrecs
        self.srec = srec
        self.offset = offset
        self.block_size = block_size

    def ensure_table(self):
        cur = self.conn.cursor()
        ensure_table(cur, "id integer primary key, data text")

    def records(self):
        for i, c in zip(range(self.nrecs), cycle(string.ascii_letters)):
            s = c * self.srec
            yield (i + self.offset, s)

    def file(self):
        f = StringIO()
        for i, s in self.records():
            f.write("%s\t%s\n" % (i, s))

        f.seek(0)
        return f

    def blocks(self):
        f = self.file()
        while block := f.read(self.block_size):
            yield block

    def assert_data(self):
        cur = self.conn.cursor()
        cur.execute("select id, data from copy_in order by id")
        for record in self.records():
            assert record == cur.fetchone()

        assert cur.fetchone() is None

    def sha(self, f):
        m = hashlib.sha256()
        while block := f.read():
            if isinstance(block, str):
                block = block.encode()
            m.update(block)
        return m.hexdigest()



"""
PostgreSQL type modifiers.

The type modifiers parse catalog information to obtain the type modifier
of a column - the numeric part of varchar(10) or decimal(6,2).
"""

# Copyright (C) 2024 The Psycopg Team

from __future__ import annotations


class TypeModifier:
    """Type modifier that doesn't know any modifier.

    Useful to describe types with no type modifier.
    """

    def __init__(self, oid: int):
        self.oid = oid

    def get_modifier(self, typemod: int) -> tuple[int, ...] | None:
        return None

    def get_display_size(self, typemod: int) -> int | None:
        return None

    def get_precision(self, typemod: int) -> int | None:
        return None

    def get_scale(self, typemod: int) -> int | None:
        return None


class NumericTypeModifier(TypeModifier):
    """Handle numeric type modifier."""

    def get_modifier(self, typemod: int) -> tuple[int, ...] | None:
        precision = self.get_precision(typemod)
        scale = self.get_scale(typemod)
        return None if precision is None or scale is None else (precision, scale)

    def get_precision(self, typemod: int) -> int | None:
        return typemod >> 16 if typemod >= 0 else None

    def get_scale(self, typemod: int) -> int | None:
        if typemod < 0:
            return None

        scale = (typemod - 4) & 0xFFFF
        if scale >= 0x400:
            scale = scale - 0x800
        return scale


class CharTypeModifier(TypeModifier):
    """Handle char/varchar type modifier."""

    def get_modifier(self, typemod: int) -> tuple[int, ...] | None:
        dsize = self.get_display_size(typemod)
        return (dsize,) if dsize else None

    def get_display_size(self, typemod: int) -> int | None:
        return typemod - 4 if typemod >= 0 else None


class BitTypeModifier(TypeModifier):
    """Handle bit/varbit type modifier."""

    def get_modifier(self, typemod: int) -> tuple[int, ...] | None:
        dsize = self.get_display_size(typemod)
        return (dsize,) if dsize else None

    def get_display_size(self, typemod: int) -> int | None:
        return typemod if typemod >= 0 else None


class TimeTypeModifier(TypeModifier):
    """Handle time-related types modifier."""

    def get_modifier(self, typemod: int) -> tuple[int, ...] | None:
        prec = self.get_precision(typemod)
        return (prec,) if prec is not None else None

    def get_precision(self, typemod: int) -> int | None:
        return typemod & 0xFFFF if typemod >= 0 else None



"""
libpq debugging tools

These functionalities are exposed here for convenience, but are not part of
the public interface and are subject to change at any moment.

Suggested usage::

    import logging
    import psycopg
    from psycopg import pq
    from psycopg.pq._debug import PGconnDebug

    logging.basicConfig(level=logging.INFO, format="%(message)s")
    logger = logging.getLogger("psycopg.debug")
    logger.setLevel(logging.INFO)

    assert pq.__impl__ == "python"
    pq.PGconn = PGconnDebug

    with psycopg.connect("") as conn:
        conn.pgconn.trace(2)
        conn.pgconn.set_trace_flags(
            pq.Trace.SUPPRESS_TIMESTAMPS | pq.Trace.REGRESS_MODE)
        ...

"""

# Copyright (C) 2022 The Psycopg Team

import inspect
import logging
from typing import Any
from functools import wraps
from collections.abc import Callable

from . import PGconn, abc
from .misc import connection_summary
from .._compat import Self, TypeVar

Func = TypeVar("Func", bound=Callable[..., Any])

logger = logging.getLogger("psycopg.debug")


class PGconnDebug:
    """Wrapper for a PQconn logging all its access."""

    _pgconn: abc.PGconn

    def __init__(self, pgconn: abc.PGconn):
        super().__setattr__("_pgconn", pgconn)

    def __repr__(self) -> str:
        cls = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
        info = connection_summary(self._pgconn)
        return f"<{cls} {info} at 0x{id(self):x}>"

    def __getattr__(self, attr: str) -> Any:
        if callable(value := getattr(self._pgconn, attr)):
            return debugging(value)
        else:
            logger.info("PGconn.%s -> %s", attr, value)
            return value

    def __setattr__(self, attr: str, value: Any) -> None:
        setattr(self._pgconn, attr, value)
        logger.info("PGconn.%s <- %s", attr, value)

    @classmethod
    def connect(cls, conninfo: bytes) -> Self:
        return cls(debugging(PGconn.connect)(conninfo))

    @classmethod
    def connect_start(cls, conninfo: bytes) -> Self:
        return cls(debugging(PGconn.connect_start)(conninfo))

    @classmethod
    def ping(self, conninfo: bytes) -> int:
        return debugging(PGconn.ping)(conninfo)


def debugging(f: Func) -> Func:
    """Wrap a function in order to log its arguments and return value on call."""

    @wraps(f)
    def debugging_(*args: Any, **kwargs: Any) -> Any:
        reprs = []
        for arg in args:
            reprs.append(f"{arg!r}")
        for k, v in kwargs.items():
            reprs.append(f"{k}={v!r}")

        logger.info("PGconn.%s(%s)", f.__name__, ", ".join(reprs))
        try:
            rv = f(*args, **kwargs)
        except Exception as ex:
            logger.info("    <- %r", ex)
            raise
        else:
            # Display the return value only if the function is declared to return
            # something else than None.
            ra = inspect.signature(f).return_annotation
            if ra is not None or rv is not None:
                logger.info("    <- %r", rv)
            return rv

    return debugging_  # type: ignore



#!/usr/bin/env python
"""Copy operation micro-benchmarks."""

from __future__ import annotations

import sys
import asyncio
import logging
from time import time
from typing import Any
from argparse import ArgumentParser, Namespace

import psycopg
from psycopg import sql
from psycopg.abc import Query

logger = logging.getLogger()
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)s %(message)s",
)


def main():
    args = parse_cmdline()
    logger.setLevel(args.loglevel)

    if getattr(args, "async"):
        asyncio.run(main_async(args))
    else:
        main_sync(args)


def main_sync(args: Namespace) -> None:
    test = CopyPutTest(args)
    with psycopg.Connection.connect(args.dsn) as conn:
        with conn.cursor() as cur:
            writer = getattr(psycopg.copy, args.writer)(cur) if args.writer else None
            cur.execute(test.get_table_stmt())
            t0 = time()
            with cur.copy(test.get_copy_stmt(), writer=writer) as copy:
                for i in range(args.nrecs):
                    copy.write_row(test.get_record())
            tf = time()

    logger.info("time to copy: %.6f sec", tf - t0)


async def main_async(args: Namespace) -> None:
    test = CopyPutTest(args)
    async with await psycopg.AsyncConnection.connect(args.dsn) as conn:
        async with conn.cursor() as cur:
            await cur.execute(test.get_table_stmt())
            writer = getattr(psycopg.copy, args.writer)(cur) if args.writer else None
            t0 = time()
            async with cur.copy(test.get_copy_stmt(), writer=writer) as copy:
                for i in range(args.nrecs):
                    await copy.write_row(test.get_record())
            tf = time()

    logger.info("time to copy: %.6f sec", tf - t0)


class CopyPutTest:
    def __init__(self, args: Namespace):
        self.args = args

    def get_table_stmt(self) -> Query:
        fields = sql.SQL(", ").join(
            [sql.SQL(f"f{i} text") for i in range(self.args.nfields)]
        )
        stmt = sql.SQL("""\
create temp table testcopy (id serial primary key, {})
""").format(fields)
        return stmt

    def get_copy_stmt(self) -> Query:
        fields = sql.SQL(", ").join(
            [sql.Identifier(f"f{i}") for i in range(self.args.nfields)]
        )
        stmt = sql.SQL("""\
copy testcopy ({}) from stdin
""").format(fields)
        return stmt

    def get_record(self) -> tuple[Any, ...]:
        return tuple("x" * self.args.colsize for _ in range(self.args.nfields))


def parse_cmdline() -> Namespace:
    parser = ArgumentParser(description=__doc__)
    parser.add_argument("--dsn", default="", help="database connection string")
    parser.add_argument(
        "--async", action="store_true", default=False, help="test async objects"
    )
    parser.add_argument(
        "--nrecs",
        type=int,
        default=1000,
        help="number of records to write [default: %(default)s]",
    )
    parser.add_argument(
        "--nfields",
        type=int,
        default=10,
        help="number of columns to write [default: %(default)s]",
    )
    parser.add_argument(
        "--colsize",
        type=int,
        default=10,
        help="width of each column to write [default: %(default)s]",
    )
    parser.add_argument("--writer", help="test alternative writer")

    g = parser.add_mutually_exclusive_group()
    g.add_argument(
        "-q",
        "--quiet",
        help="Talk less",
        dest="loglevel",
        action="store_const",
        const=logging.WARN,
        default=logging.INFO,
    )
    g.add_argument(
        "-v",
        "--verbose",
        help="Talk more",
        dest="loglevel",
        action="store_const",
        const=logging.DEBUG,
        default=logging.INFO,
    )

    if (args := parser.parse_args()).writer:
        try:
            getattr(psycopg.copy, args.writer)
        except AttributeError:
            parser.error(f"unknown writer: {args.writer!r}")

    return args


if __name__ == "__main__":
    sys.exit(main())



import sys
import json
import subprocess as sp

import pytest

import psycopg

pytest.importorskip("gevent")

pytestmark = [pytest.mark.gevent]


@pytest.mark.slow
@pytest.mark.timing
def test_gevent(dsn):
    TICK = 0.1
    script = f"""\
import gevent.monkey
gevent.monkey.patch_all()

import json
import time
import gevent
import psycopg

TICK = {TICK!r}
dts = []
queried = False

def ticker():
    t0 = time.time()
    for i in range(5):
        time.sleep(TICK)
        t = time.time()
        dts.append(t - t0)
        t0 = t

def querier():
    time.sleep(TICK * 2)
    with psycopg.connect({dsn!r}) as conn:
        conn.execute("select pg_sleep(0.3)")

    global queried
    queried = True

jobs = [gevent.spawn(ticker), gevent.spawn(querier)]
gevent.joinall(jobs, timeout=3)
print(json.dumps(dts))
"""
    cmdline = [sys.executable, "-c", script]
    rv = sp.run(cmdline, check=True, text=True, stdout=sp.PIPE)
    dts = json.loads(rv.stdout)

    for dt in dts:
        assert TICK <= dt < TICK * 1.2


@pytest.mark.skipif("not psycopg._cmodule._psycopg")
def test_patched_dont_use_wait_c():
    if psycopg.waiting.wait is not psycopg.waiting.wait_c:
        pytest.skip("wait_c not normally in use")

    script = """
import gevent.monkey
gevent.monkey.patch_all()

import psycopg
assert psycopg.waiting.wait is not psycopg.waiting.wait_c
"""
    sp.check_call([sys.executable, "-c", script])


@pytest.mark.skipif("not psycopg._cmodule._psycopg")
def test_unpatched_still_use_wait_c():
    if psycopg.waiting.wait is not psycopg.waiting.wait_c:
        pytest.skip("wait_c not normally in use")

    script = """
import gevent.monkey

import psycopg
assert psycopg.waiting.wait is psycopg.waiting.wait_c
"""
    sp.check_call([sys.executable, "-c", script])



"""
Adapters for None.
"""

# Copyright (C) 2020 The Psycopg Team

from __future__ import annotations

from ..abc import AdaptContext, Buffer, NoneType
from ..adapt import Dumper


class NoneDumper(Dumper):
    """
    Not a complete dumper as it doesn't implement dump(), but it implements
    quote(), so it can be used in sql composition.
    """

    def dump(self, obj: None) -> Buffer | None:
        raise NotImplementedError("NULL is passed to Postgres in other ways")

    def quote(self, obj: None) -> Buffer:
        return b"NULL"


def register_default_adapters(context: AdaptContext) -> None:
    context.adapters.register_dumper(NoneType, NoneDumper)



"""
C optimized functions to manipulate arrays
"""

# Copyright (C) 2022 The Psycopg Team

import cython

from cpython.mem cimport PyMem_Free, PyMem_Realloc
from cpython.ref cimport Py_INCREF
from libc.stdint cimport int32_t, uint32_t
from libc.string cimport strchr
from cpython.list cimport PyList_Append, PyList_GET_ITEM, PyList_GET_SIZE, PyList_New
from cpython.list cimport PyList_SET_ITEM
from cpython.object cimport PyObject

from psycopg_c._psycopg cimport endian
from psycopg_c.pq.libpq cimport Oid

from psycopg import errors as e


cdef extern from *:
    """
/* Defined in PostgreSQL in src/include/utils/array.h */
#define MAXDIM 6
    """
    const int MAXDIM


cdef class ArrayLoader(_CRecursiveLoader):

    format = PQ_TEXT
    base_oid = 0
    delimiter = b","

    cdef PyObject *row_loader
    cdef char cdelim

    # A memory area which used to unescape elements.
    # Keep it here to avoid a malloc per element and to set up exceptions
    # to make sure to free it on error.
    cdef char *scratch
    cdef size_t sclen

    cdef object cload(self, const char *data, size_t length):
        if self.cdelim == b"\x00":
            self.row_loader = self._tx._c_get_loader(
                <PyObject *>self.base_oid, <PyObject *>PQ_TEXT)
            self.cdelim = self.delimiter[0]

        return _array_load_text(
            data, length, self.row_loader, self.cdelim,
            &(self.scratch), &(self.sclen))

    def __dealloc__(self):
        PyMem_Free(self.scratch)


@cython.final
cdef class ArrayBinaryLoader(_CRecursiveLoader):

    format = PQ_BINARY

    cdef PyObject *row_loader

    cdef object cload(self, const char *data, size_t length):
        rv = _array_load_binary(data, length, self._tx, &(self.row_loader))
        return rv


cdef object _array_load_text(
    const char *buf, size_t length, PyObject *row_loader, char cdelim,
    char **scratch, size_t *sclen
):
    if length == 0:
        raise e.DataError("malformed array: empty data")

    cdef const char *end = buf + length

    # Remove the dimensions information prefix (``[...]=``)
    if buf[0] == b"[":
        buf = strchr(buf + 1, b'=')
        if buf == NULL:
            raise e.DataError("malformed array: no '=' after dimension information")
        buf += 1

    # TODO: further optimization: pre-scan the array to find the array
    # dimensions, so that we can preallocate the list sized instead of calling
    # append, which is the dominating operation

    cdef list stack = []
    cdef list a = []
    rv = a
    cdef PyObject *tmp

    cdef CLoader cloader = None
    cdef object pyload = None
    if (<RowLoader>row_loader).cloader is not None:
        cloader = (<RowLoader>row_loader).cloader
    else:
        pyload = (<RowLoader>row_loader).loadfunc

    while buf < end:
        if buf[0] == b'{':
            if stack:
                tmp = PyList_GET_ITEM(stack, PyList_GET_SIZE(stack) - 1)
                PyList_Append(<object>tmp, a)
            PyList_Append(stack, a)
            a = []
            buf += 1

        elif buf[0] == b'}':
            if not stack:
                raise e.DataError("malformed array: unexpected '}'")
            rv = stack.pop()
            buf += 1

        elif buf[0] == cdelim:
            buf += 1

        else:
            v = _parse_token(
                &buf, end, cdelim, scratch, sclen, cloader, pyload)
            if not stack:
                raise e.DataError("malformed array: missing initial '{'")
            tmp = PyList_GET_ITEM(stack, PyList_GET_SIZE(stack) - 1)
            PyList_Append(<object>tmp, v)

    return rv


cdef object _parse_token(
    const char **bufptr, const char *bufend, char cdelim,
    char **scratch, size_t *sclen, CLoader cloader, object load
):
    cdef const char *start = bufptr[0]
    cdef int has_quotes = start[0] == b'"'
    cdef int quoted = has_quotes
    cdef int num_escapes = 0
    cdef int escaped = 0

    if has_quotes:
        start += 1
    cdef const char *end = start

    while end < bufend:
        if (end[0] == cdelim or end[0] == b'}') and not quoted:
            break
        elif end[0] == b'\\' and not escaped:
            num_escapes += 1
            escaped = 1
            end += 1
            continue
        elif end[0] == b'"' and not escaped:
            quoted = 0
        escaped = 0
        end += 1
    else:
        raise e.DataError("malformed array: hit the end of the buffer")

    # Return the new position for the buffer
    bufptr[0] = end
    if has_quotes:
        end -= 1

    cdef Py_ssize_t length = (end - start)
    if length == 4 and not has_quotes \
            and start[0] == b'N' and start[1] == b'U' \
            and start[2] == b'L' and start[3] == b'L':
        return None

    cdef const char *src
    cdef char *tgt
    cdef size_t unesclen

    if not num_escapes:
        if cloader is not None:
            return cloader.cload(start, length)
        else:
            b = start[:length]
            return load(b)

    else:
        unesclen = length - num_escapes + 1
        if unesclen > sclen[0]:
            scratch[0] = <char *>PyMem_Realloc(scratch[0], unesclen)
            sclen[0] = unesclen

        src = start
        tgt = scratch[0]
        while src < end:
            if src[0] == b'\\':
                src += 1
            tgt[0] = src[0]
            src += 1
            tgt += 1

        tgt[0] = b'\x00'

        if cloader is not None:
            return cloader.cload(scratch[0], length - num_escapes)
        else:
            b = scratch[0][:length - num_escapes]
            return load(b)


@cython.cdivision(True)
cdef object _array_load_binary(
    const char *buf, size_t length, Transformer tx, PyObject **row_loader_ptr
):
    # head is ndims, hasnull, elem oid
    cdef uint32_t buf32
    memcpy(&buf32, buf, sizeof(buf32))
    cdef int ndims = endian.be32toh(buf32)

    if ndims <= 0:
        return []
    elif ndims > MAXDIM:
        raise e.DataError(
            r"unexpected number of dimensions %s exceeding the maximum allowed %s"
            % (ndims, MAXDIM)
        )

    cdef object oid
    cdef uint32_t beoid
    if row_loader_ptr[0] == NULL:
        memcpy(&beoid, buf + 2 * sizeof(uint32_t), sizeof(beoid))
        oid = <Oid>endian.be32toh(beoid)
        row_loader_ptr[0] = tx._c_get_loader(<PyObject *>oid, <PyObject *>PQ_BINARY)

    cdef Py_ssize_t[MAXDIM] dims
    cdef int i
    cdef uint32_t bedata
    cdef const char *dimptr = buf + 3 * sizeof(uint32_t)
    for i in range(ndims):
        # Every dimension is dim, lower bound
        memcpy(&bedata, dimptr, sizeof(bedata))
        dimptr += 2 * sizeof(uint32_t)
        dims[i] = endian.be32toh(bedata)

    buf += (3 + 2 * ndims) * sizeof(bedata)
    out = _array_load_binary_rec(ndims, dims, &buf, row_loader_ptr[0])
    return out


cdef object _array_load_binary_rec(
    Py_ssize_t ndims, Py_ssize_t *dims, const char **bufptr, PyObject *row_loader
):
    cdef const char *buf
    cdef int i
    cdef uint32_t besize
    cdef int32_t size
    cdef object val

    cdef Py_ssize_t nelems = dims[0]
    cdef list out = PyList_New(nelems)

    if ndims == 1:
        buf = bufptr[0]
        for i in range(nelems):
            memcpy(&besize, buf, sizeof(besize))
            size = <int32_t>endian.be32toh(besize)
            buf += sizeof(besize)
            if size == -1:
                val = None
            else:
                if (<RowLoader>row_loader).cloader is not None:
                    val = (<RowLoader>row_loader).cloader.cload(buf, size)
                else:
                    val = (<RowLoader>row_loader).loadfunc(buf[:size])
                buf += size

            Py_INCREF(val)
            PyList_SET_ITEM(out, i, val)

        bufptr[0] = buf

    else:
        for i in range(nelems):
            val = _array_load_binary_rec(ndims - 1, dims + 1, bufptr, row_loader)
            Py_INCREF(val)
            PyList_SET_ITEM(out, i, val)

    return out



"""
compatibility functions for different Python versions
"""

# Copyright (C) 2021 The Psycopg Team

from __future__ import annotations

import sys
from typing import Any, Iterator

if sys.version_info >= (3, 11):
    from typing import LiteralString, Self
else:
    from typing_extensions import LiteralString, Self

if sys.version_info >= (3, 12):
    _asyncio_run_snippet = (
        "running 'asyncio.run(...,"
        " loop_factory=asyncio.SelectorEventLoop(selectors.SelectSelector()))'"
    )

else:
    _asyncio_run_snippet = (
        "setting 'asyncio.set_event_loop_policy(WindowsSelectorEventLoopPolicy())'"
    )

if sys.version_info >= (3, 13):
    from typing import TypeVar
else:
    from typing_extensions import TypeVar

if sys.version_info >= (3, 14):
    from string.templatelib import Interpolation, Template
else:
    from dataclasses import dataclass

    class Template:
        strings: tuple[str]
        interpolations: tuple[Interpolation]

        def __new__(cls, *args: str | Interpolation) -> Self:
            return cls()

        def __iter__(self) -> Iterator[str | Interpolation]:
            return
            yield

    @dataclass
    class Interpolation:
        value: Any
        expression: str
        conversion: str | None
        format_spec: str


__all__ = [
    "Interpolation",
    "LiteralString",
    "Self",
    "Template",
    "TypeVar",
]



"""
types stub for ctypes functions
"""

# Copyright (C) 2020 The Psycopg Team

from ctypes import Array, _Pointer, c_char, c_char_p, c_int, c_ubyte, c_uint, c_ulong
from ctypes import pointer
from typing import Any, Callable, Sequence

class FILE: ...

def fdopen(fd: int, mode: bytes) -> _Pointer[FILE]: ...  # type: ignore[type-var]

Oid = c_uint

class PGconn_struct: ...
class PGresult_struct: ...
class PGcancelConn_struct: ...
class PGcancel_struct: ...

class PQconninfoOption_struct:
    keyword: bytes
    envvar: bytes
    compiled: bytes
    val: bytes
    label: bytes
    dispchar: bytes
    dispsize: int

class PGnotify_struct:
    be_pid: int
    relname: bytes
    extra: bytes

class PGresAttDesc_struct:
    name: bytes
    tableid: int
    columnid: int
    format: int
    typid: int
    typlen: int
    atttypmod: int

def PQhostaddr(arg1: PGconn_struct | None) -> bytes: ...
def PQerrorMessage(arg1: PGconn_struct | None) -> bytes: ...
def PQresultErrorMessage(arg1: PGresult_struct | None) -> bytes: ...
def PQexecPrepared(
    arg1: PGconn_struct | None,
    arg2: bytes,
    arg3: int,
    arg4: Array[c_char_p] | None,
    arg5: Array[c_int] | None,
    arg6: Array[c_int] | None,
    arg7: int,
) -> PGresult_struct: ...
def PQprepare(
    arg1: PGconn_struct | None,
    arg2: bytes,
    arg3: bytes,
    arg4: int,
    arg5: Array[c_uint] | None,
) -> PGresult_struct: ...
def PQgetvalue(
    arg1: PGresult_struct | None, arg2: int, arg3: int
) -> _Pointer[c_char]: ...
def PQcmdTuples(arg1: PGresult_struct | None) -> bytes: ...
def PQescapeStringConn(
    arg1: PGconn_struct | None,
    arg2: c_char_p,
    arg3: bytes,
    arg4: int,
    arg5: _Pointer[c_int],
) -> int: ...
def PQescapeString(arg1: c_char_p, arg2: bytes, arg3: int) -> int: ...
def PQsendPrepare(
    arg1: PGconn_struct | None,
    arg2: bytes,
    arg3: bytes,
    arg4: int,
    arg5: Array[c_uint] | None,
) -> int: ...
def PQsendQueryPrepared(
    arg1: PGconn_struct | None,
    arg2: bytes,
    arg3: int,
    arg4: Array[c_char_p] | None,
    arg5: Array[c_int] | None,
    arg6: Array[c_int] | None,
    arg7: int,
) -> int: ...
def PQcancelCreate(arg1: PGconn_struct | None) -> PGcancelConn_struct: ...
def PQcancelBlocking(arg1: PGcancelConn_struct | None) -> int: ...
def PQcancelStart(arg1: PGcancelConn_struct | None) -> int: ...
def PQcancelPoll(arg1: PGcancelConn_struct | None) -> int: ...
def PQcancelStatus(arg1: PGcancelConn_struct | None) -> int: ...
def PQcancelSocket(arg1: PGcancelConn_struct | None) -> int: ...
def PQcancelErrorMessage(arg1: PGcancelConn_struct | None) -> bytes: ...
def PQcancelReset(arg1: PGcancelConn_struct | None) -> None: ...
def PQcancelFinish(arg1: PGcancelConn_struct | None) -> None: ...
def PQcancel(arg1: PGcancel_struct | None, arg2: c_char_p, arg3: int) -> int: ...
def PQsetNoticeReceiver(
    arg1: PGconn_struct, arg2: Callable[[Any], PGresult_struct], arg3: Any
) -> Callable[[Any], PGresult_struct]: ...

# TODO: Ignoring type as getting an error on mypy/ctypes:
# Type argument "psycopg.pq._pq_ctypes.PGnotify_struct" of "pointer" must be
# a subtype of "ctypes._CData"
def PQnotifies(
    arg1: PGconn_struct | None,
) -> _Pointer[PGnotify_struct] | None: ...  # type: ignore
def PQputCopyEnd(arg1: PGconn_struct | None, arg2: bytes | None) -> int: ...

# Arg 2 is a _Pointer, reported as _CArgObject by mypy
def PQgetCopyData(arg1: PGconn_struct | None, arg2: Any, arg3: int) -> int: ...
def PQsetResultAttrs(
    arg1: PGresult_struct | None,
    arg2: int,
    arg3: Array[PGresAttDesc_struct],  # type: ignore
) -> int: ...
def PQtrace(
    arg1: PGconn_struct | None,
    arg2: _Pointer[FILE],  # type: ignore[type-var]
) -> None: ...
def PQsetTraceFlags(arg1: PGconn_struct | None, arg2: int) -> None: ...
def PQencryptPasswordConn(
    arg1: PGconn_struct | None,
    arg2: bytes,
    arg3: bytes,
    arg4: bytes | None,
) -> bytes: ...
def PQpipelineStatus(pgconn: PGconn_struct | None) -> int: ...
def PQenterPipelineMode(pgconn: PGconn_struct | None) -> int: ...
def PQexitPipelineMode(pgconn: PGconn_struct | None) -> int: ...
def PQpipelineSync(pgconn: PGconn_struct | None) -> int: ...
def PQsendFlushRequest(pgconn: PGconn_struct | None) -> int: ...

# Autogenerated section.
# In order to refresh, run:
#   python -m psycopg.pq._pq_ctypes

# fmt: off
# autogenerated: start
def PQlibVersion() -> int: ...
def PQconnectdb(arg1: bytes) -> PGconn_struct: ...
def PQconnectStart(arg1: bytes) -> PGconn_struct: ...
def PQconnectPoll(arg1: PGconn_struct | None) -> int: ...
def PQconndefaults() -> Sequence[PQconninfoOption_struct]: ...
def PQconninfoFree(arg1: Sequence[PQconninfoOption_struct]) -> None: ...
def PQconninfo(arg1: PGconn_struct | None) -> Sequence[PQconninfoOption_struct]: ...
def PQconninfoParse(arg1: bytes, arg2: _Pointer[c_char_p]) -> Sequence[PQconninfoOption_struct]: ...
def PQfinish(arg1: PGconn_struct | None) -> None: ...
def PQreset(arg1: PGconn_struct | None) -> None: ...
def PQresetStart(arg1: PGconn_struct | None) -> int: ...
def PQresetPoll(arg1: PGconn_struct | None) -> int: ...
def PQping(arg1: bytes) -> int: ...
def PQdb(arg1: PGconn_struct | None) -> bytes | None: ...
def PQuser(arg1: PGconn_struct | None) -> bytes | None: ...
def PQpass(arg1: PGconn_struct | None) -> bytes | None: ...
def PQhost(arg1: PGconn_struct | None) -> bytes | None: ...
def PQport(arg1: PGconn_struct | None) -> bytes | None: ...
def PQtty(arg1: PGconn_struct | None) -> bytes | None: ...
def PQoptions(arg1: PGconn_struct | None) -> bytes | None: ...
def PQstatus(arg1: PGconn_struct | None) -> int: ...
def PQtransactionStatus(arg1: PGconn_struct | None) -> int: ...
def PQparameterStatus(arg1: PGconn_struct | None, arg2: bytes) -> bytes | None: ...
def PQprotocolVersion(arg1: PGconn_struct | None) -> int: ...
def PQfullProtocolVersion(arg1: PGconn_struct | None) -> int: ...
def PQserverVersion(arg1: PGconn_struct | None) -> int: ...
def PQsocket(arg1: PGconn_struct | None) -> int: ...
def PQbackendPID(arg1: PGconn_struct | None) -> int: ...
def PQconnectionNeedsPassword(arg1: PGconn_struct | None) -> int: ...
def PQconnectionUsedPassword(arg1: PGconn_struct | None) -> int: ...
def PQconnectionUsedGSSAPI(arg1: PGconn_struct | None) -> int: ...
def PQsslInUse(arg1: PGconn_struct | None) -> int: ...
def PQexec(arg1: PGconn_struct | None, arg2: bytes) -> PGresult_struct: ...
def PQexecParams(arg1: PGconn_struct | None, arg2: bytes, arg3: int, arg4: _Pointer[c_uint], arg5: _Pointer[c_char_p], arg6: _Pointer[c_int], arg7: _Pointer[c_int], arg8: int) -> PGresult_struct: ...
def PQdescribePrepared(arg1: PGconn_struct | None, arg2: bytes) -> PGresult_struct: ...
def PQdescribePortal(arg1: PGconn_struct | None, arg2: bytes) -> PGresult_struct: ...
def PQclosePrepared(arg1: PGconn_struct | None, arg2: bytes) -> PGresult_struct: ...
def PQclosePortal(arg1: PGconn_struct | None, arg2: bytes) -> PGresult_struct: ...
def PQresultStatus(arg1: PGresult_struct | None) -> int: ...
def PQresultErrorField(arg1: PGresult_struct | None, arg2: int) -> bytes | None: ...
def PQclear(arg1: PGresult_struct | None) -> None: ...
def PQntuples(arg1: PGresult_struct | None) -> int: ...
def PQnfields(arg1: PGresult_struct | None) -> int: ...
def PQfname(arg1: PGresult_struct | None, arg2: int) -> bytes | None: ...
def PQftable(arg1: PGresult_struct | None, arg2: int) -> int: ...
def PQftablecol(arg1: PGresult_struct | None, arg2: int) -> int: ...
def PQfformat(arg1: PGresult_struct | None, arg2: int) -> int: ...
def PQftype(arg1: PGresult_struct | None, arg2: int) -> int: ...
def PQfmod(arg1: PGresult_struct | None, arg2: int) -> int: ...
def PQfsize(arg1: PGresult_struct | None, arg2: int) -> int: ...
def PQbinaryTuples(arg1: PGresult_struct | None) -> int: ...
def PQgetisnull(arg1: PGresult_struct | None, arg2: int, arg3: int) -> int: ...
def PQgetlength(arg1: PGresult_struct | None, arg2: int, arg3: int) -> int: ...
def PQnparams(arg1: PGresult_struct | None) -> int: ...
def PQparamtype(arg1: PGresult_struct | None, arg2: int) -> int: ...
def PQcmdStatus(arg1: PGresult_struct | None) -> bytes | None: ...
def PQoidValue(arg1: PGresult_struct | None) -> int: ...
def PQescapeLiteral(arg1: PGconn_struct | None, arg2: bytes, arg3: int) -> bytes | None: ...
def PQescapeIdentifier(arg1: PGconn_struct | None, arg2: bytes, arg3: int) -> bytes | None: ...
def PQescapeByteaConn(arg1: PGconn_struct | None, arg2: bytes, arg3: int, arg4: _Pointer[c_ulong]) -> _Pointer[c_ubyte]: ...
def PQescapeBytea(arg1: bytes, arg2: int, arg3: _Pointer[c_ulong]) -> _Pointer[c_ubyte]: ...
def PQunescapeBytea(arg1: bytes, arg2: _Pointer[c_ulong]) -> _Pointer[c_ubyte]: ...
def PQsendQuery(arg1: PGconn_struct | None, arg2: bytes) -> int: ...
def PQsendQueryParams(arg1: PGconn_struct | None, arg2: bytes, arg3: int, arg4: _Pointer[c_uint], arg5: _Pointer[c_char_p], arg6: _Pointer[c_int], arg7: _Pointer[c_int], arg8: int) -> int: ...
def PQsendDescribePrepared(arg1: PGconn_struct | None, arg2: bytes) -> int: ...
def PQsendDescribePortal(arg1: PGconn_struct | None, arg2: bytes) -> int: ...
def PQsendClosePrepared(arg1: PGconn_struct | None, arg2: bytes) -> int: ...
def PQsendClosePortal(arg1: PGconn_struct | None, arg2: bytes) -> int: ...
def PQgetResult(arg1: PGconn_struct | None) -> PGresult_struct: ...
def PQconsumeInput(arg1: PGconn_struct | None) -> int: ...
def PQisBusy(arg1: PGconn_struct | None) -> int: ...
def PQsetnonblocking(arg1: PGconn_struct | None, arg2: int) -> int: ...
def PQisnonblocking(arg1: PGconn_struct | None) -> int: ...
def PQflush(arg1: PGconn_struct | None) -> int: ...
def PQsetSingleRowMode(arg1: PGconn_struct | None) -> int: ...
def PQsetChunkedRowsMode(arg1: PGconn_struct | None, arg2: int) -> int: ...
def PQgetCancel(arg1: PGconn_struct | None) -> PGcancel_struct: ...
def PQfreeCancel(arg1: PGcancel_struct | None) -> None: ...
def PQputCopyData(arg1: PGconn_struct | None, arg2: bytes, arg3: int) -> int: ...
def PQuntrace(arg1: PGconn_struct | None) -> None: ...
def PQfreemem(arg1: Any) -> None: ...
def PQchangePassword(arg1: PGconn_struct | None, arg2: bytes, arg3: bytes) -> PGresult_struct: ...
def PQmakeEmptyPGresult(arg1: PGconn_struct | None, arg2: int) -> PGresult_struct: ...
def PQinitOpenSSL(arg1: int, arg2: int) -> None: ...
# autogenerated: end
# fmt: on



"""
Adapters for textual types.
"""

# Copyright (C) 2020 The Psycopg Team

from __future__ import annotations

from typing import TYPE_CHECKING

from .. import _oids
from ..pq import Escaping, Format
from ..abc import AdaptContext
from ..adapt import Buffer, Dumper, Loader
from ..errors import DataError
from .._encodings import conn_encoding

if TYPE_CHECKING:
    from ..pq.abc import Escaping as EscapingProto


class _BaseStrDumper(Dumper):
    def __init__(self, cls: type, context: AdaptContext | None = None):
        super().__init__(cls, context)
        enc = conn_encoding(self.connection)
        self._encoding = enc if enc != "ascii" else "utf-8"


class _StrBinaryDumper(_BaseStrDumper):
    """
    Base class to dump a Python strings to a Postgres text type, in binary format.

    Subclasses shall specify the oids of real types (text, varchar, name...).
    """

    format = Format.BINARY

    def dump(self, obj: str) -> Buffer | None:
        # the server will raise DataError subclass if the string contains 0x00
        return obj.encode(self._encoding)


class _StrDumper(_BaseStrDumper):
    """
    Base class to dump a Python strings to a Postgres text type, in text format.

    Subclasses shall specify the oids of real types (text, varchar, name...).
    """

    def dump(self, obj: str) -> Buffer | None:
        if "\x00" in obj:
            raise DataError("PostgreSQL text fields cannot contain NUL (0x00) bytes")
        else:
            return obj.encode(self._encoding)


# The next are concrete dumpers, each one specifying the oid they dump to.


class StrBinaryDumper(_StrBinaryDumper):
    oid = _oids.TEXT_OID


class StrBinaryDumperVarchar(_StrBinaryDumper):
    oid = _oids.VARCHAR_OID


class StrBinaryDumperName(_StrBinaryDumper):
    oid = _oids.NAME_OID


class StrDumper(_StrDumper):
    """
    Dumper for strings in text format to the text oid.

    Note that this dumper is not used by default because the type is too strict
    and PostgreSQL would require an explicit casts to everything that is not a
    text field. However it is useful where the unknown oid is ambiguous and the
    text oid is required, for instance with variadic functions.
    """

    oid = _oids.TEXT_OID


class StrDumperVarchar(_StrDumper):
    oid = _oids.VARCHAR_OID


class StrDumperName(_StrDumper):
    oid = _oids.NAME_OID


class StrDumperUnknown(_StrDumper):
    """
    Dumper for strings in text format to the unknown oid.

    This dumper is the default dumper for strings and allows to use Python
    strings to represent almost every data type. In a few places, however, the
    unknown oid is not accepted (for instance in variadic functions such as
    'concat()'). In that case either a cast on the placeholder ('%s::text') or
    the StrTextDumper should be used.
    """

    pass


class TextLoader(Loader):
    def __init__(self, oid: int, context: AdaptContext | None = None):
        super().__init__(oid, context)
        enc = conn_encoding(self.connection)
        self._encoding = enc if enc != "ascii" else ""

    def load(self, data: Buffer) -> bytes | str:
        if self._encoding:
            if isinstance(data, memoryview):
                data = bytes(data)
            return data.decode(self._encoding)
        else:
            # return bytes for SQL_ASCII db
            if not isinstance(data, bytes):
                data = bytes(data)
            return data


class TextBinaryLoader(TextLoader):
    format = Format.BINARY


class BytesDumper(Dumper):
    oid = _oids.BYTEA_OID
    _qprefix = b""

    def __init__(self, cls: type, context: AdaptContext | None = None):
        super().__init__(cls, context)
        self._esc = Escaping(self.connection.pgconn if self.connection else None)

    def dump(self, obj: Buffer) -> Buffer | None:
        return self._esc.escape_bytea(obj)

    def quote(self, obj: Buffer) -> Buffer:
        if (escaped := self.dump(obj)) is None:
            return b"NULL"

        # We cannot use the base quoting because escape_bytea already returns
        # the quotes content. if scs is off it will escape the backslashes in
        # the format, otherwise it won't, but it doesn't tell us what quotes to
        # use.
        if self.connection:
            if not self._qprefix:
                scs = self.connection.pgconn.parameter_status(
                    b"standard_conforming_strings"
                )
                self._qprefix = b"'" if scs == b"on" else b" E'"

            return self._qprefix + escaped + b"'"

        # We don't have a connection, so someone is using us to generate a file
        # to use off-line or something like that. PQescapeBytea, like its
        # string counterpart, is not predictable whether it will escape
        # backslashes.
        rv: bytes = b" E'" + escaped + b"'"
        if self._esc.escape_bytea(b"\x00") == b"\\000":
            rv = rv.replace(b"\\", b"\\\\")
        return rv


class BytesBinaryDumper(Dumper):
    format = Format.BINARY
    oid = _oids.BYTEA_OID

    def dump(self, obj: Buffer) -> Buffer | None:
        return obj


class ByteaLoader(Loader):
    _escaping: EscapingProto

    def __init__(self, oid: int, context: AdaptContext | None = None):
        super().__init__(oid, context)
        if not hasattr(self.__class__, "_escaping"):
            self.__class__._escaping = Escaping()

    def load(self, data: Buffer) -> bytes:
        return self._escaping.unescape_bytea(data)


class ByteaBinaryLoader(Loader):
    format = Format.BINARY

    def load(self, data: Buffer) -> Buffer:
        return data


def register_default_adapters(context: AdaptContext) -> None:
    adapters = context.adapters

    # NOTE: the order the dumpers are registered is relevant. The last one
    # registered becomes the default for each type. Usually, binary is the
    # default dumper. For text we use the text dumper as default because it
    # plays the role of unknown, and it can be cast automatically to other
    # types. However, before that, we register dumper with 'text', 'varchar',
    # 'name' oids, which will be used when a text dumper is looked up by oid.
    adapters.register_dumper(str, StrBinaryDumperName)
    adapters.register_dumper(str, StrBinaryDumperVarchar)
    adapters.register_dumper(str, StrBinaryDumper)
    adapters.register_dumper(str, StrDumperName)
    adapters.register_dumper(str, StrDumperVarchar)
    adapters.register_dumper(str, StrDumper)
    adapters.register_dumper(str, StrDumperUnknown)

    adapters.register_loader(_oids.INVALID_OID, TextLoader)
    adapters.register_loader("bpchar", TextLoader)
    adapters.register_loader("name", TextLoader)
    adapters.register_loader("text", TextLoader)
    adapters.register_loader("varchar", TextLoader)
    adapters.register_loader('"char"', TextLoader)
    adapters.register_loader("bpchar", TextBinaryLoader)
    adapters.register_loader("name", TextBinaryLoader)
    adapters.register_loader("text", TextBinaryLoader)
    adapters.register_loader("varchar", TextBinaryLoader)
    adapters.register_loader('"char"', TextBinaryLoader)

    adapters.register_dumper(bytes, BytesDumper)
    adapters.register_dumper(bytearray, BytesDumper)
    adapters.register_dumper(memoryview, BytesDumper)
    adapters.register_dumper(bytes, BytesBinaryDumper)
    adapters.register_dumper(bytearray, BytesBinaryDumper)
    adapters.register_dumper(memoryview, BytesBinaryDumper)

    adapters.register_loader("bytea", ByteaLoader)
    adapters.register_loader(_oids.INVALID_OID, ByteaBinaryLoader)
    adapters.register_loader("bytea", ByteaBinaryLoader)



from __future__ import annotations

import gc
import sys
import sysconfig

import pytest


def pytest_collection_modifyitems(items):
    for item in items:
        if "gc" in item.fixturenames:
            item.add_marker(pytest.mark.refcount)


def pytest_configure(config):
    config.addinivalue_line(
        "markers",
        "refcount: the test checks ref counts which is sometimes flaky",
    )


NO_COUNT_TYPES: tuple[type, ...] = ()

if sys.version_info[:2] == (3, 10):
    # On my laptop there are occasional creations of a single one of these objects
    # with empty content, which might be some Decimal caching.
    # Keeping the guard as strict as possible, to be extended if other types
    # or versions are necessary.
    try:
        from _contextvars import Context
    except ImportError:
        pass
    else:
        NO_COUNT_TYPES += (Context,)


class GCFixture:
    __slots__ = ()

    @staticmethod
    def collect() -> None:
        """
        gc.collect(), but more insisting.
        """
        for i in range(3):
            gc.collect()

    @staticmethod
    def count() -> int:
        """
        len(gc.get_objects()), with subtleties.
        """

        if not NO_COUNT_TYPES:
            return len(gc.get_objects())

        # Note: not using a list comprehension because it pollutes the objects list.
        rv = 0
        for obj in gc.get_objects():
            if isinstance(obj, NO_COUNT_TYPES):
                continue
            rv += 1

        return rv


@pytest.fixture(name="gc")
def fixture_gc():
    """
    Provides a consistent way to run garbage collection and count references.

    **Note:** This will skip tests on PyPy.
    """
    if sys.implementation.name == "pypy" or sysconfig.get_config_var("Py_GIL_DISABLED"):
        pytest.skip(reason="depends on refcount semantics")
    return GCFixture()


@pytest.fixture
def gc_collect():
    """
    Provides a consistent way to run garbage collection.

    **Note:** This will *not* skip tests on PyPy.
    """
    if sysconfig.get_config_var("Py_GIL_DISABLED"):
        pytest.skip(reason="depends on refcount semantics")
    return GCFixture.collect



import ipaddress

import pytest

from psycopg import pq, sql
from psycopg.adapt import PyFormat

crdb_skip_inet = pytest.mark.crdb_skip("inet")
crdb_skip_cidr = pytest.mark.crdb_skip("cidr")


@crdb_skip_inet
@pytest.mark.parametrize("fmt_in", PyFormat)
@pytest.mark.parametrize("val", ["192.168.0.1", "2001:db8::"])
def test_address_dump(conn, fmt_in, val):
    cur = conn.cursor()
    cur.execute(f"select %{fmt_in.value} = %s::inet", (ipaddress.ip_address(val), val))
    assert cur.fetchone()[0] is True
    cur.execute(
        f"select %{fmt_in.value} = array[null, %s]::inet[]",
        ([None, ipaddress.ip_interface(val)], val),
    )
    assert cur.fetchone()[0] is True


@crdb_skip_inet
@pytest.mark.parametrize("fmt_in", PyFormat)
@pytest.mark.parametrize("val", ["127.0.0.1/24", "::ffff:102:300/128"])
def test_interface_dump(conn, fmt_in, val):
    cur = conn.cursor()
    rec = cur.execute(
        f"select %(val){fmt_in.value} = %(repr)s::inet,"
        f" %(val){fmt_in.value}, %(repr)s::inet",
        {"val": ipaddress.ip_interface(val), "repr": val},
    ).fetchone()
    assert rec[0] is True, f"{rec[1]} != {rec[2]}"
    cur.execute(
        f"select %{fmt_in.value} = array[null, %s]::inet[]",
        ([None, ipaddress.ip_interface(val)], val),
    )
    assert cur.fetchone()[0] is True


@crdb_skip_cidr
@pytest.mark.parametrize("fmt_in", PyFormat)
@pytest.mark.parametrize("val", ["127.0.0.0/24", "::ffff:102:300/128"])
def test_network_dump(conn, fmt_in, val):
    cur = conn.cursor()
    cur.execute(f"select %{fmt_in.value} = %s::cidr", (ipaddress.ip_network(val), val))
    assert cur.fetchone()[0] is True
    cur.execute(
        f"select %{fmt_in.value} = array[NULL, %s]::cidr[]",
        ([None, ipaddress.ip_network(val)], val),
    )
    assert cur.fetchone()[0] is True


@crdb_skip_cidr
@pytest.mark.parametrize("fmt_in", PyFormat)
def test_network_mixed_size_array(conn, fmt_in):
    val = [
        ipaddress.IPv4Network("192.168.0.1/32"),
        ipaddress.IPv6Network("::1/128"),
    ]
    cur = conn.cursor()
    cur.execute(f"select %{fmt_in.value}", (val,))
    got = cur.fetchone()[0]
    assert val == got


@pytest.mark.crdb_skip("copy")
@pytest.mark.parametrize("fmt_out", pq.Format)
@pytest.mark.parametrize("val", ["127.0.0.1/32", "::ffff:102:300/128"])
def test_inet_load_address(conn, fmt_out, val):
    addr = ipaddress.ip_address(val.split("/", 1)[0])
    cur = conn.cursor(binary=fmt_out)

    cur.execute("select %s::inet", (val,))
    assert cur.fetchone()[0] == addr

    cur.execute("select array[null, %s::inet]", (val,))
    assert cur.fetchone()[0] == [None, addr]

    stmt = sql.SQL("copy (select {}::inet) to stdout (format {})").format(
        val, sql.SQL(fmt_out.name)
    )
    with cur.copy(stmt) as copy:
        copy.set_types(["inet"])
        (got,) = copy.read_row()

    assert got == addr


@pytest.mark.crdb_skip("copy")
@pytest.mark.parametrize("fmt_out", pq.Format)
@pytest.mark.parametrize("val", ["127.0.0.1/24", "::ffff:102:300/127"])
def test_inet_load_network(conn, fmt_out, val):
    pyval = ipaddress.ip_interface(val)
    cur = conn.cursor(binary=fmt_out)

    cur.execute("select %s::inet", (val,))
    assert cur.fetchone()[0] == pyval

    cur.execute("select array[null, %s::inet]", (val,))
    assert cur.fetchone()[0] == [None, pyval]

    stmt = sql.SQL("copy (select {}::inet) to stdout (format {})").format(
        val, sql.SQL(fmt_out.name)
    )
    with cur.copy(stmt) as copy:
        copy.set_types(["inet"])
        (got,) = copy.read_row()

    assert got == pyval


@crdb_skip_cidr
@pytest.mark.parametrize("fmt_out", pq.Format)
@pytest.mark.parametrize("val", ["127.0.0.0/24", "::ffff:102:300/128"])
def test_cidr_load(conn, fmt_out, val):
    pyval = ipaddress.ip_network(val)
    cur = conn.cursor(binary=fmt_out)

    cur.execute("select %s::cidr", (val,))
    assert cur.fetchone()[0] == pyval

    cur.execute("select array[null, %s::cidr]", (val,))
    assert cur.fetchone()[0] == [None, pyval]

    stmt = sql.SQL("copy (select {}::cidr) to stdout (format {})").format(
        val, sql.SQL(fmt_out.name)
    )
    with cur.copy(stmt) as copy:
        copy.set_types(["cidr"])
        (got,) = copy.read_row()

    assert got == pyval



"""
psycopg two-phase commit support
"""

# Copyright (C) 2021 The Psycopg Team

from __future__ import annotations

import re
import datetime as dt
from base64 import b64decode, b64encode
from dataclasses import dataclass, replace

_re_xid = re.compile(r"^(\d+)_([^_]*)_([^_]*)$")


@dataclass(frozen=True)
class Xid:
    """A two-phase commit transaction identifier.

    The object can also be unpacked as a 3-item tuple (`format_id`, `gtrid`,
    `bqual`).

    """

    format_id: int | None
    gtrid: str
    bqual: str | None
    prepared: dt.datetime | None = None
    owner: str | None = None
    database: str | None = None

    @classmethod
    def from_string(cls, s: str) -> Xid:
        """Try to parse an XA triple from the string.

        This may fail for several reasons. In such case return an unparsed Xid.
        """
        try:
            return cls._parse_string(s)
        except Exception:
            return Xid(None, s, None)

    def __str__(self) -> str:
        return self._as_tid()

    def __len__(self) -> int:
        return 3

    def __getitem__(self, index: int) -> int | str | None:
        return (self.format_id, self.gtrid, self.bqual)[index]

    @classmethod
    def _parse_string(cls, s: str) -> Xid:
        if not (m := _re_xid.match(s)):
            raise ValueError("bad Xid format")

        format_id = int(m.group(1))
        gtrid = b64decode(m.group(2)).decode()
        bqual = b64decode(m.group(3)).decode()
        return cls.from_parts(format_id, gtrid, bqual)

    @classmethod
    def from_parts(cls, format_id: int | None, gtrid: str, bqual: str | None) -> Xid:
        if format_id is not None:
            if bqual is None:
                raise TypeError("if format_id is specified, bqual must be too")
            if not 0 <= format_id < 0x80000000:
                raise ValueError("format_id must be a non-negative 32-bit integer")
            if len(bqual) > 64:
                raise ValueError("bqual must be not longer than 64 chars")
            if len(gtrid) > 64:
                raise ValueError("gtrid must be not longer than 64 chars")

        elif bqual is None:
            raise TypeError("if format_id is None, bqual must be None too")

        return Xid(format_id, gtrid, bqual)

    def _as_tid(self) -> str:
        """
        Return the PostgreSQL transaction_id for this XA xid.

        PostgreSQL wants just a string, while the DBAPI supports the XA
        standard and thus a triple. We use the same conversion algorithm
        implemented by JDBC in order to allow some form of interoperation.

        see also: the pgjdbc implementation
          http://cvs.pgfoundry.org/cgi-bin/cvsweb.cgi/jdbc/pgjdbc/org/
            postgresql/xa/RecoveredXid.java?rev=1.2
        """
        if self.format_id is None or self.bqual is None:
            # Unparsed xid: return the gtrid.
            return self.gtrid

        # XA xid: mash together the components.
        egtrid = b64encode(self.gtrid.encode()).decode()
        ebqual = b64encode(self.bqual.encode()).decode()

        return f"{self.format_id}_{egtrid}_{ebqual}"

    @classmethod
    def _get_recover_query(cls) -> str:
        return "SELECT gid, prepared, owner, database FROM pg_prepared_xacts"

    @classmethod
    def _from_record(
        cls, gid: str, prepared: dt.datetime, owner: str, database: str
    ) -> Xid:
        xid = Xid.from_string(gid)
        return replace(xid, prepared=prepared, owner=owner, database=database)


Xid.__module__ = "psycopg"



import pickle

import pytest

from psycopg.postgres import types as builtins

from .fix_crdb import crdb_encoding, is_crdb, skip_crdb


def test_description_attribs(conn):
    curs = conn.cursor()
    curs.execute("""select
        3.14::decimal(10,2) as pi,
        'hello'::text as hi,
        '2010-02-18'::date as now
        """)
    assert len(curs.description) == 3
    for c in curs.description:
        len(c) == 7  # DBAPI happy
        for i, a in enumerate("""
            name type_code display_size internal_size precision scale null_ok
        """.split()):
            assert c[i] == getattr(c, a)

        # Won't fill them up
        assert c.null_ok is None

    c = curs.description[0]
    assert c.name == "pi"
    assert c.type_code == builtins["numeric"].oid
    assert c.display_size is None
    assert c.internal_size is None
    assert c.precision == 10
    assert c.scale == 2

    c = curs.description[1]
    assert c.name == "hi"
    assert c.type_code == builtins["text"].oid
    assert c.display_size is None
    assert c.internal_size is None
    assert c.precision is None
    assert c.scale is None

    c = curs.description[2]
    assert c.name == "now"
    assert c.type_code == builtins["date"].oid
    assert c.display_size is None
    if is_crdb(conn) and conn.info.server_version < 230000:
        assert c.internal_size == 16
    else:
        assert c.internal_size == 4
    assert c.precision is None
    assert c.scale is None


def test_description_slice(conn):
    curs = conn.cursor()
    curs.execute("select 1::int as a")
    curs.description[0][0:2] == ("a", 23)


def skip_neg_scale(*args):
    return pytest.param(
        *args,
        marks=[
            pytest.mark.crdb("skip", reason="negative precision numeric"),
            pytest.mark.pg(">=15"),
        ],
    )


@pytest.mark.parametrize(
    "type, precision, scale, dsize, isize",
    [
        ("text", None, None, None, None),
        ("text[]", None, None, None, None),
        ("varchar", None, None, None, None),
        ("varchar(1)", None, None, 1, None),
        ("varchar(1)[]", None, None, 1, None),
        ("varchar(42)", None, None, 42, None),
        skip_crdb("bpchar(42)", None, None, 42, None, reason="bpchar"),
        ("varchar(10485760)", None, None, 10485760, None),
        ("int4", None, None, None, 4),
        ("numeric", None, None, None, None),
        ("numeric(10,0)", 10, 0, None, None),
        ("numeric(10,3)[]", 10, 3, None, None),
        skip_neg_scale("numeric(2,-3)", 2, -3, None, None),
        skip_neg_scale("numeric(3,5)", 3, 5, None, None),
        skip_neg_scale("numeric(1,-1000)", 1, -1000, None, None),
        skip_neg_scale("numeric(1,1000)", 1, 1000, None, None),
        ("numeric(1000,1000)", 1000, 1000, None, None),
        ("time", None, None, None, 8),
        ("time[]", None, None, None, None),
        ("timetz", None, None, None, 12),
        ("timestamp", None, None, None, 8),
        ("timestamptz", None, None, None, 8),
        ("interval", None, None, None, 16),
        ("bit(1)", None, None, 1, None),
        ("bit(42)", None, None, 42, None),
        ("bit(83886080)", None, None, 83886080, None),
        ("varbit", None, None, None, None),
        ("varbit(1)", None, None, 1, None),
        ("varbit(42)", None, None, 42, None),
        ("varbit(83886080)", None, None, 83886080, None),
    ],
)
def test_details(conn, type, precision, scale, dsize, isize):
    cur = conn.cursor()
    cur.execute(f"select null::{type}")
    col = cur.description[0]
    assert type == col.type_display
    assert f" {type} " in (repr(col))
    assert col.precision == precision
    assert col.scale == scale
    assert col.display_size == dsize
    if not is_crdb(conn):
        assert col.internal_size == isize


@pytest.mark.crdb("skip", reason="time precision")
@pytest.mark.parametrize("type", "time timetz timestamp timestamptz interval".split())
@pytest.mark.parametrize("precision", [0, 2, 6])
def test_details_time(conn, type, precision):
    type = f"{type}({precision})"
    cur = conn.cursor()
    cur.execute(f"select null::{type}")
    col = cur.description[0]
    assert type in (repr(col))
    assert col.precision == precision


def test_pickle(conn):
    curs = conn.cursor()
    curs.execute("""select
        3.14::decimal(10,2) as pi,
        'hello'::text as hi,
        '2010-02-18'::date as now
        """)
    description = curs.description
    pickled = pickle.dumps(description, pickle.HIGHEST_PROTOCOL)
    unpickled = pickle.loads(pickled)
    assert [tuple(d) for d in description] == [tuple(d) for d in unpickled]


@pytest.mark.crdb_skip("no col query")
def test_no_col_query(conn):
    cur = conn.execute("select")
    assert cur.description == []
    assert cur.fetchall() == [()]


def test_description_closed_connection(conn):
    # If we have reasons to break this test we will (e.g. we really need
    # the connection). In #172 it fails just by accident.
    cur = conn.execute("select 1::int4 as foo")
    conn.close()
    assert len(cur.description) == 1
    col = cur.description[0]
    assert col.name == "foo"
    assert col.type_code == 23


def test_name_not_a_name(conn):
    cur = conn.cursor()
    (res,) = cur.execute("""select 'x' as "foo-bar" """).fetchone()
    assert res == "x"
    assert cur.description[0].name == "foo-bar"


@pytest.mark.parametrize("encoding", ["utf8", crdb_encoding("latin9")])
def test_name_encode(conn, encoding):
    conn.execute(f"set client_encoding to {encoding}")
    cur = conn.cursor()
    (res,) = cur.execute("""select 'x' as "\u20ac" """).fetchone()
    assert res == "x"
    assert cur.description[0].name == "\u20ac"



# type: ignore  # dnspython is currently optional and mypy fails if missing
"""
DNS query support
"""

# Copyright (C) 2021 The Psycopg Team

from __future__ import annotations

import os
import re
import warnings
from random import randint
from typing import TYPE_CHECKING, Any, DefaultDict, NamedTuple
from collections import defaultdict
from collections.abc import Sequence

try:
    from dns.resolver import Cache, Resolver
    from dns.exception import DNSException
    from dns.asyncresolver import Resolver as AsyncResolver
except ImportError:
    raise ImportError(
        "the module psycopg._dns requires the package 'dnspython' installed"
    )

from . import conninfo
from . import errors as e

if TYPE_CHECKING:
    from dns.rdtypes.IN.SRV import SRV

resolver = Resolver()
resolver.cache = Cache()

async_resolver = AsyncResolver()
async_resolver.cache = Cache()


async def resolve_hostaddr_async(params: dict[str, Any]) -> dict[str, Any]:
    """
    Perform async DNS lookup of the hosts and return a new params dict.

    .. deprecated:: 3.1
        The use of this function is not necessary anymore, because
        `psycopg.AsyncConnection.connect()` performs non-blocking name
        resolution automatically.
    """
    warnings.warn(
        "from psycopg 3.1, resolve_hostaddr_async() is not needed anymore",
        DeprecationWarning,
    )
    hosts: list[str] = []
    hostaddrs: list[str] = []
    ports: list[str] = []

    for attempt in await conninfo.conninfo_attempts_async(params):
        if attempt.get("host") is not None:
            hosts.append(attempt["host"])
        if attempt.get("hostaddr") is not None:
            hostaddrs.append(attempt["hostaddr"])
        if attempt.get("port") is not None:
            ports.append(str(attempt["port"]))

    out = params.copy()
    if shosts := ",".join(hosts):
        out["host"] = shosts
    if shostaddrs := ",".join(hostaddrs):
        out["hostaddr"] = shostaddrs
    sports = ",".join(ports)
    if ports:
        out["port"] = sports

    return out


def resolve_srv(params: dict[str, Any]) -> dict[str, Any]:
    """Apply SRV DNS lookup as defined in :RFC:`2782`."""
    return Rfc2782Resolver().resolve(params)


async def resolve_srv_async(params: dict[str, Any]) -> dict[str, Any]:
    """Async equivalent of `resolve_srv()`."""
    return await Rfc2782Resolver().resolve_async(params)


class HostPort(NamedTuple):
    host: str
    port: str
    totry: bool = False
    target: str | None = None


class Rfc2782Resolver:
    """Implement SRV RR Resolution as per RFC 2782

    The class is organised to minimise code duplication between the sync and
    the async paths.
    """

    re_srv_rr = re.compile(r"^(?P<service>_[^\.]+)\.(?P<proto>_[^\.]+)\.(?P<target>.+)")

    def resolve(self, params: dict[str, Any]) -> dict[str, Any]:
        """Update the parameters host and port after SRV lookup."""
        if not (attempts := self._get_attempts(params)):
            return params

        hps = []
        for hp in attempts:
            if hp.totry:
                hps.extend(self._resolve_srv(hp))
            else:
                hps.append(hp)

        return self._return_params(params, hps)

    async def resolve_async(self, params: dict[str, Any]) -> dict[str, Any]:
        """Update the parameters host and port after SRV lookup."""
        if not (attempts := self._get_attempts(params)):
            return params

        hps = []
        for hp in attempts:
            if hp.totry:
                hps.extend(await self._resolve_srv_async(hp))
            else:
                hps.append(hp)

        return self._return_params(params, hps)

    def _get_attempts(self, params: dict[str, Any]) -> list[HostPort]:
        """
        Return the list of host, and for each host if SRV lookup must be tried.

        Return an empty list if no lookup is requested.
        """
        # If hostaddr is defined don't do any resolution.
        if params.get("hostaddr", os.environ.get("PGHOSTADDR", "")):
            return []

        host_arg: str = params.get("host", os.environ.get("PGHOST", ""))
        hosts_in = host_arg.split(",")
        port_arg: str = str(params.get("port", os.environ.get("PGPORT", "")))
        if len(ports_in := port_arg.split(",")) == 1:
            # If only one port is specified, it applies to all the hosts.
            ports_in *= len(hosts_in)
        if len(ports_in) != len(hosts_in):
            # ProgrammingError would have been more appropriate, but this is
            # what the raise if the libpq fails connect in the same case.
            raise e.OperationalError(
                f"cannot match {len(hosts_in)} hosts with {len(ports_in)} port numbers"
            )

        out = []
        srv_found = False
        for host, port in zip(hosts_in, ports_in):
            if (m := self.re_srv_rr.match(host)) or port.lower() == "srv":
                srv_found = True
                target = m.group("target") if m else None
                hp = HostPort(host=host, port=port, totry=True, target=target)
            else:
                hp = HostPort(host=host, port=port)
            out.append(hp)

        return out if srv_found else []

    def _resolve_srv(self, hp: HostPort) -> list[HostPort]:
        try:
            ans = resolver.resolve(hp.host, "SRV")
        except DNSException:
            ans = ()
        return self._get_solved_entries(hp, ans)

    async def _resolve_srv_async(self, hp: HostPort) -> list[HostPort]:
        try:
            ans = await async_resolver.resolve(hp.host, "SRV")
        except DNSException:
            ans = ()
        return self._get_solved_entries(hp, ans)

    def _get_solved_entries(
        self, hp: HostPort, entries: Sequence[SRV]
    ) -> list[HostPort]:
        if not entries:
            # No SRV entry found. Delegate the libpq a QNAME=target lookup
            if hp.target and hp.port.lower() != "srv":
                return [HostPort(host=hp.target, port=hp.port)]
            else:
                return []

        # If there is precisely one SRV RR, and its Target is "." (the root
        # domain), abort.
        if len(entries) == 1 and str(entries[0].target) == ".":
            return []

        return [
            HostPort(host=str(entry.target).rstrip("."), port=str(entry.port))
            for entry in self.sort_rfc2782(entries)
        ]

    def _return_params(
        self, params: dict[str, Any], hps: list[HostPort]
    ) -> dict[str, Any]:
        if not hps:
            # Nothing found, we ended up with an empty list
            raise e.OperationalError("no host found after SRV RR lookup")

        out = params.copy()
        out["host"] = ",".join(hp.host for hp in hps)
        out["port"] = ",".join(str(hp.port) for hp in hps)
        return out

    def sort_rfc2782(self, ans: Sequence[SRV]) -> list[SRV]:
        """
        Implement the priority/weight ordering defined in RFC 2782.
        """
        # Divide the entries by priority:
        priorities: DefaultDict[int, list[SRV]] = defaultdict(list)
        out: list[SRV] = []
        for entry in ans:
            priorities[entry.priority].append(entry)

        for pri, entries in sorted(priorities.items()):
            if len(entries) == 1:
                out.append(entries[0])
                continue

            entries.sort(key=lambda ent: ent.weight)
            total_weight = sum(ent.weight for ent in entries)
            while entries:
                r = randint(0, total_weight)
                csum = 0
                for i, ent in enumerate(entries):
                    csum += ent.weight
                    if csum >= r:
                        break
                out.append(ent)
                total_weight -= ent.weight
                del entries[i]

        return out



from __future__ import annotations

import enum
from math import exp, isinf, isnan
from decimal import Decimal

import pytest

import psycopg
from psycopg import pq, sql
from psycopg.abc import Buffer
from psycopg.adapt import PyFormat, Transformer
from psycopg.types.numeric import FloatLoader, Int8, Int8BinaryDumper, Int8Dumper

from ..fix_crdb import is_crdb

#
# Tests with int
#


@pytest.mark.parametrize(
    "val, expr",
    [
        (0, "'0'::int"),
        (1, "'1'::int"),
        (-1, "'-1'::int"),
        (42, "'42'::smallint"),
        (-42, "'-42'::smallint"),
        (int(2**63 - 1), "'9223372036854775807'::bigint"),
        (int(-(2**63)), "'-9223372036854775808'::bigint"),
    ],
)
@pytest.mark.parametrize("fmt_in", PyFormat)
def test_dump_int(conn, val, expr, fmt_in):
    assert isinstance(val, int)
    cur = conn.cursor()
    cur.execute(f"select {expr} = %{fmt_in.value}", (val,))
    assert cur.fetchone()[0] is True


@pytest.mark.parametrize(
    "val, expr",
    [
        (0, "'0'::smallint"),
        (1, "'1'::smallint"),
        (-1, "'-1'::smallint"),
        (42, "'42'::smallint"),
        (-42, "'-42'::smallint"),
        (int(2**15 - 1), f"'{2 ** 15 - 1}'::smallint"),
        (int(-(2**15)), f"'{-2 ** 15}'::smallint"),
        (int(2**15), f"'{2 ** 15}'::integer"),
        (int(-(2**15) - 1), f"'{-2 ** 15 - 1}'::integer"),
        (int(2**31 - 1), f"'{2 ** 31 - 1}'::integer"),
        (int(-(2**31)), f"'{-2 ** 31}'::integer"),
        (int(2**31), f"'{2 ** 31}'::bigint"),
        (int(-(2**31) - 1), f"'{-2 ** 31 - 1}'::bigint"),
        (int(2**63 - 1), f"'{2 ** 63 - 1}'::bigint"),
        (int(-(2**63)), f"'{-2 ** 63}'::bigint"),
        (int(2**63), f"'{2 ** 63}'::numeric"),
        (int(-(2**63) - 1), f"'{-2 ** 63 - 1}'::numeric"),
    ],
)
@pytest.mark.parametrize("fmt_in", PyFormat)
def test_dump_int_subtypes(conn, val, expr, fmt_in):
    cur = conn.cursor()
    cur.execute(f"select pg_typeof({expr}) = pg_typeof(%{fmt_in.value})", (val,))
    assert cur.fetchone()[0] is True
    cur.execute(
        f"select {expr} = %(v){fmt_in.value}, {expr}::text, %(v){fmt_in.value}::text",
        {"v": val},
    )
    ok, want, got = cur.fetchone()
    assert got == want
    assert ok


@pytest.mark.parametrize("fmt_in", [PyFormat.TEXT, PyFormat.BINARY])
def test_int_none(conn, fmt_in):
    Base: type = Int8Dumper if fmt_in == PyFormat.TEXT else Int8BinaryDumper

    class MyDumper(Base):  # type: ignore
        def dump(self, obj: int) -> Buffer | None:
            if not obj:
                return None
            else:
                return super().dump(obj)  # type: ignore

    conn.adapters.register_dumper(Int8, MyDumper)
    cur = conn.execute("select %s, %s", [Int8(0), Int8(1)])
    assert cur.fetchone() == (None, 1)


class MyEnum(enum.IntEnum):
    foo = 42


class MyMixinEnum(int, enum.Enum):
    foo = 42000000


@pytest.mark.parametrize("fmt_in", PyFormat)
@pytest.mark.parametrize("enum", [MyEnum, MyMixinEnum])
def test_dump_enum(conn, fmt_in, enum):
    cur = conn.cursor()
    cur.execute(f"select %{fmt_in.value}", (enum.foo,))
    (res,) = cur.fetchone()
    assert res == enum.foo.value


@pytest.mark.parametrize(
    "val, expr",
    [
        (0, b"0"),
        (1, b"1"),
        (-1, b" -1"),
        (42, b"42"),
        (-42, b" -42"),
        (int(2**63 - 1), b"9223372036854775807"),
        (int(-(2**63)), b" -9223372036854775808"),
        (int(2**63), b"9223372036854775808"),
        (int(-(2**63 + 1)), b" -9223372036854775809"),
        (int(2**100), b"1267650600228229401496703205376"),
        (int(-(2**100)), b" -1267650600228229401496703205376"),
    ],
)
def test_quote_int(conn, val, expr):
    tx = Transformer()
    assert tx.get_dumper(val, PyFormat.TEXT).quote(val) == expr

    cur = conn.cursor()
    cur.execute(sql.SQL("select {v}, -{v}").format(v=sql.Literal(val)))
    assert cur.fetchone() == (val, -val)


@pytest.mark.parametrize(
    "val, pgtype, want",
    [
        ("0", "integer", 0),
        ("1", "integer", 1),
        ("-1", "integer", -1),
        ("0", "int2", 0),
        ("0", "int4", 0),
        ("0", "int8", 0),
        ("0", "integer", 0),
        ("0", "oid", 0),
        # bounds
        ("-32768", "smallint", -32768),
        ("+32767", "smallint", 32767),
        ("-2147483648", "integer", -2147483648),
        ("+2147483647", "integer", 2147483647),
        ("-9223372036854775808", "bigint", -9223372036854775808),
        ("9223372036854775807", "bigint", 9223372036854775807),
        ("4294967295", "oid", 4294967295),
    ],
)
@pytest.mark.parametrize("fmt_out", pq.Format)
def test_load_int(conn, val, pgtype, want, fmt_out):
    if pgtype == "integer" and is_crdb(conn):
        pgtype = "int4"  # "integer" is "int8" on crdb
    cur = conn.cursor(binary=fmt_out)
    cur.execute(f"select %s::{pgtype}", (val,))
    assert cur.pgresult.fformat(0) == fmt_out
    assert cur.pgresult.ftype(0) == conn.adapters.types[pgtype].oid
    result = cur.fetchone()[0]
    assert result == want
    assert type(result) is type(want)

    # arrays work too
    cur.execute(f"select array[%s::{pgtype}]", (val,))
    assert cur.pgresult.fformat(0) == fmt_out
    assert cur.pgresult.ftype(0) == conn.adapters.types[pgtype].array_oid
    result = cur.fetchone()[0]
    assert result == [want]
    assert type(result[0]) is type(want)


#
# Tests with float
#


@pytest.mark.parametrize(
    "val, expr",
    [
        (0.0, "'0'"),
        (1.0, "'1'"),
        (-1.0, "'-1'"),
        (float("nan"), "'NaN'"),
        (float("inf"), "'Infinity'"),
        (float("-inf"), "'-Infinity'"),
    ],
)
@pytest.mark.parametrize("fmt_in", PyFormat)
def test_dump_float(conn, val, expr, fmt_in):
    assert isinstance(val, float)
    cur = conn.cursor()
    cur.execute(f"select %{fmt_in.value} = {expr}::float8", (val,))
    assert cur.fetchone()[0] is True


@pytest.mark.parametrize(
    "val, expr",
    [
        (0.0, b"0.0"),
        (1.0, b"1.0"),
        (10000000000000000.0, b"1e+16"),
        (1000000.1, b"1000000.1"),
        (-100000.000001, b" -100000.000001"),
        (-1.0, b" -1.0"),
        (float("nan"), b"'NaN'::float8"),
        (float("inf"), b"'Infinity'::float8"),
        (float("-inf"), b"'-Infinity'::float8"),
    ],
)
def test_quote_float(conn, val, expr):
    tx = Transformer()
    assert tx.get_dumper(val, PyFormat.TEXT).quote(val) == expr

    cur = conn.cursor()
    cur.execute(sql.SQL("select {v}, -{v}").format(v=sql.Literal(val)))
    r = cur.fetchone()
    if isnan(val):
        assert isnan(r[0]) and isnan(r[1])
    else:
        if isinstance(r[0], Decimal):
            r = tuple(map(float, r))

        assert r == (val, -val)


@pytest.mark.parametrize(
    "val, expr",
    [
        (exp(1), "exp(1.0)"),
        (-exp(1), "-exp(1.0)"),
        (1e30, "'1e30'"),
        (1e-30, "1e-30"),
        (-1e30, "'-1e30'"),
        (-1e-30, "-1e-30"),
    ],
)
def test_dump_float_approx(conn, val, expr):
    assert isinstance(val, float)
    cur = conn.cursor()
    cur.execute(f"select abs(({expr}::float8 - %s) / {expr}::float8) <= 1e-15", (val,))
    assert cur.fetchone()[0] is True

    cur.execute(f"select abs(({expr}::float4 - %s) / {expr}::float4) <= 1e-6", (val,))
    assert cur.fetchone()[0] is True


@pytest.mark.parametrize(
    "val, pgtype, want",
    [
        ("0", "float4", 0.0),
        ("0.0", "float4", 0.0),
        ("42", "float4", 42.0),
        ("-42", "float4", -42.0),
        ("0.0", "float8", 0.0),
        ("0.0", "real", 0.0),
        ("0.0", "double precision", 0.0),
        ("0.0", "float4", 0.0),
        ("nan", "float4", float("nan")),
        ("inf", "float4", float("inf")),
        ("-inf", "float4", -float("inf")),
        ("nan", "float8", float("nan")),
        ("inf", "float8", float("inf")),
        ("-inf", "float8", -float("inf")),
    ],
)
@pytest.mark.parametrize("fmt_out", pq.Format)
def test_load_float(conn, val, pgtype, want, fmt_out):
    cur = conn.cursor(binary=fmt_out)
    cur.execute(f"select %s::{pgtype}", (val,))
    assert cur.pgresult.fformat(0) == fmt_out
    assert cur.pgresult.ftype(0) == conn.adapters.types[pgtype].oid
    result = cur.fetchone()[0]

    def check(result, want):
        assert type(result) is type(want)
        if isnan(want):
            assert isnan(result)
        elif isinf(want):
            assert isinf(result)
            assert (result < 0) is (want < 0)
        else:
            assert result == want

    check(result, want)

    cur.execute(f"select array[%s::{pgtype}]", (val,))
    assert cur.pgresult.fformat(0) == fmt_out
    assert cur.pgresult.ftype(0) == conn.adapters.types[pgtype].array_oid
    result = cur.fetchone()[0]
    assert isinstance(result, list)
    check(result[0], want)


@pytest.mark.parametrize(
    "expr, pgtype, want",
    [
        ("exp(1.0)", "float4", 2.71828),
        ("-exp(1.0)", "float4", -2.71828),
        ("exp(1.0)", "float8", 2.71828182845905),
        ("-exp(1.0)", "float8", -2.71828182845905),
        ("1.42e10", "float4", 1.42e10),
        ("-1.42e10", "float4", -1.42e10),
        ("1.42e40", "float8", 1.42e40),
        ("-1.42e40", "float8", -1.42e40),
    ],
)
@pytest.mark.parametrize("fmt_out", pq.Format)
def test_load_float_approx(conn, expr, pgtype, want, fmt_out):
    cur = conn.cursor(binary=fmt_out)
    cur.execute("select %s::%s" % (expr, pgtype))
    assert cur.pgresult.fformat(0) == fmt_out
    result = cur.fetchone()[0]
    assert result == pytest.approx(want)


@pytest.mark.crdb_skip("copy")
def test_load_float_copy(conn):
    cur = conn.cursor(binary=False)
    with cur.copy("copy (select 3.14::float8, 'hi'::text) to stdout;") as copy:
        copy.set_types(["float8", "text"])
        rec = copy.read_row()

    assert rec[0] == pytest.approx(3.14)
    assert rec[1] == "hi"


#
# Tests with decimal
#


@pytest.mark.parametrize(
    "val",
    [
        "0",
        "-0",
        "0.0",
        "0.000000000000000000001",
        "-0.000000000000000000001",
        "nan",
        "snan",
    ],
)
@pytest.mark.parametrize("fmt_in", PyFormat)
@pytest.mark.parametrize("fmt_out", pq.Format)
def test_roundtrip_numeric(conn, val, fmt_in, fmt_out):
    cur = conn.cursor(binary=fmt_out)
    val = Decimal(val)
    cur.execute(f"select %{fmt_in.value}", (val,))
    result = cur.fetchone()[0]
    assert isinstance(result, Decimal)
    if val.is_nan():
        assert result.is_nan()
    else:
        assert result == val


@pytest.mark.parametrize(
    "val, expr",
    [
        ("0", b"0"),
        ("0.0", b"0.0"),
        ("0.00000000000000001", b"1E-17"),
        ("-0.00000000000000001", b" -1E-17"),
        ("nan", b"'NaN'::numeric"),
        ("snan", b"'NaN'::numeric"),
    ],
)
def test_quote_numeric(conn, val, expr):
    val = Decimal(val)
    tx = Transformer()
    assert tx.get_dumper(val, PyFormat.TEXT).quote(val) == expr

    cur = conn.cursor()
    cur.execute(sql.SQL("select {v}, -{v}").format(v=sql.Literal(val)))
    r = cur.fetchone()

    if val.is_nan():
        assert isnan(r[0]) and isnan(r[1])
    else:
        assert r == (val, -val)


@pytest.mark.crdb_skip("binary decimal")
@pytest.mark.parametrize(
    "expr",
    ["NaN", "1", "1.0", "-1", "0.0", "0.01", "11", "1.1", "1.01", "0", "0.00"]
    + [
        "0.0000000",
        "0.00001",
        "1.00001",
        "-1.00000000000000",
        "-2.00000000000000",
        "1000000000.12345",
        "100.123456790000000000000000",
        "1.0e-1000",
        "1e1000",
        "0.000000000000000000000000001",
        "1.0000000000000000000000001",
        "1000000000000000000000000.001",
        "1000000000000000000000000000.001",
        "9999999999999999999999999999.9",
    ],
)
def test_dump_numeric_binary(conn, expr):
    cur = conn.cursor()
    val = Decimal(expr)
    cur.execute("select %b::text, %s::decimal::text", [val, expr])
    want, got = cur.fetchone()
    assert got == want


@pytest.mark.slow
@pytest.mark.parametrize(
    "fmt_in",
    [
        (
            f
            if f != PyFormat.BINARY
            else pytest.param(f, marks=pytest.mark.crdb_skip("binary decimal"))
        )
        for f in PyFormat
    ],
)
def test_dump_numeric_exhaustive(conn, fmt_in):
    cur = conn.cursor()

    funcs = [
        (lambda i: "1" + "0" * i),
        (lambda i: "1" + "0" * i + "." + "0" * i),
        (lambda i: "-1" + "0" * i),
        (lambda i: "0." + "0" * i + "1"),
        (lambda i: "-0." + "0" * i + "1"),
        (lambda i: "1." + "0" * i + "1"),
        (lambda i: "1." + "0" * i + "10"),
        (lambda i: "1" + "0" * i + ".001"),
        (lambda i: "9" + "9" * i),
        (lambda i: "9" + "." + "9" * i),
        (lambda i: "9" + "9" * i + ".9"),
        (lambda i: "9" + "9" * i + "." + "9" * i),
        (lambda i: "1.1e%s" % i),
        (lambda i: "1.1e-%s" % i),
    ]

    for i in range(100):
        for f in funcs:
            expr = f(i)
            val = Decimal(expr)
            cur.execute(f"select %{fmt_in.value}::text, %s::decimal::text", [val, expr])
            got, want = cur.fetchone()
            assert got == want


@pytest.mark.pg(">= 14")
@pytest.mark.parametrize("val, expr", [("inf", "Infinity"), ("-inf", "-Infinity")])
def test_dump_numeric_binary_inf(conn, val, expr):
    cur = conn.cursor()
    val = Decimal(val)
    cur.execute("select %b", [val])


@pytest.mark.parametrize(
    "expr",
    ["nan", "0", "1", "-1", "0.0", "0.01"]
    + [
        "0.0000000",
        "-1.00000000000000",
        "-2.00000000000000",
        "1000000000.12345",
        "100.123456790000000000000000",
        "1.0e-1000",
        "1e1000",
        "0.000000000000000000000000001",
        "1.0000000000000000000000001",
        "1000000000000000000000000.001",
        "1000000000000000000000000000.001",
        "9999999999999999999999999999.9",
    ],
)
def test_load_numeric_binary(conn, expr):
    cur = conn.cursor(binary=1)
    res = cur.execute(f"select '{expr}'::numeric").fetchone()[0]

    if (val := Decimal(expr)).is_nan():
        assert res.is_nan()
    else:
        assert res == val
        if "e" not in expr:
            assert str(res) == str(val)


@pytest.mark.slow
@pytest.mark.parametrize("fmt_out", pq.Format)
def test_load_numeric_exhaustive(conn, fmt_out):
    cur = conn.cursor(binary=fmt_out)

    funcs = [
        (lambda i: "1" + "0" * i),
        (lambda i: "1" + "0" * i + "." + "0" * i),
        (lambda i: "-1" + "0" * i),
        (lambda i: "0." + "0" * i + "1"),
        (lambda i: "-0." + "0" * i + "1"),
        (lambda i: "1." + "0" * i + "1"),
        (lambda i: "1." + "0" * i + "10"),
        (lambda i: "1" + "0" * i + ".001"),
        (lambda i: "9" + "9" * i),
        (lambda i: "9" + "." + "9" * i),
        (lambda i: "9" + "9" * i + ".9"),
        (lambda i: "9" + "9" * i + "." + "9" * i),
    ]

    for i in range(100):
        for f in funcs:
            snum = f(i)
            want = Decimal(snum)
            got = cur.execute(f"select '{snum}'::decimal").fetchone()[0]
            assert want == got
            assert str(want) == str(got)


@pytest.mark.pg(">= 14")
@pytest.mark.parametrize("val, expr", [("inf", "Infinity"), ("-inf", "-Infinity")])
def test_load_numeric_binary_inf(conn, val, expr):
    cur = conn.cursor(binary=1)
    res = cur.execute(f"select '{expr}'::numeric").fetchone()[0]
    val = Decimal(val)
    assert res == val


@pytest.mark.parametrize(
    "val", ["0", "0.0", "0.000000000000000000001", "-0.000000000000000000001", "nan"]
)
def test_numeric_as_float(conn, val):
    cur = conn.cursor()
    cur.adapters.register_loader("numeric", FloatLoader)

    val = Decimal(val)
    cur.execute("select %s as val", (val,))
    result = cur.fetchone()[0]
    assert isinstance(result, float)
    if val.is_nan():
        assert isnan(result)
    else:
        assert result == pytest.approx(float(val))

    # the customization works with arrays too
    cur.execute("select %s as arr", ([val],))
    result = cur.fetchone()[0]
    assert isinstance(result, list)
    assert isinstance(result[0], float)
    if val.is_nan():
        assert isnan(result[0])
    else:
        assert result[0] == pytest.approx(float(val))


#
# Mixed tests
#


@pytest.mark.parametrize("pgtype", [None, "float8", "int8", "numeric"])
def test_minus_minus(conn, pgtype):
    cur = conn.cursor()
    cast = f"::{pgtype}" if pgtype is not None else ""
    cur.execute(f"select -%s{cast}", [-1])
    result = cur.fetchone()[0]
    assert result == 1


@pytest.mark.parametrize("pgtype", [None, "float8", "int8", "numeric"])
def test_minus_minus_quote(conn, pgtype):
    cur = conn.cursor()
    cast = f"::{pgtype}" if pgtype is not None else ""
    cur.execute(sql.SQL("select -{}{}").format(sql.Literal(-1), sql.SQL(cast)))
    result = cur.fetchone()[0]
    assert result == 1


@pytest.mark.parametrize("wrapper", "Int2 Int4 Int8 Oid Float4 Float8".split())
@pytest.mark.parametrize("fmt_in", PyFormat)
def test_dump_wrapper(conn, wrapper, fmt_in):
    wrapper = getattr(psycopg.types.numeric, wrapper)
    obj = wrapper(1)
    cur = conn.execute(
        f"select %(obj){fmt_in.value} = 1, %(obj){fmt_in.value}", {"obj": obj}
    )
    rec = cur.fetchone()
    assert rec[0], rec[1]


@pytest.mark.parametrize("wrapper", "Int2 Int4 Int8 Oid Float4 Float8".split())
def test_dump_wrapper_oid(wrapper):
    wrapper = getattr(psycopg.types.numeric, wrapper)
    base = wrapper.__mro__[1]
    assert base in (int, float)
    n = base(3.14)
    assert str(wrapper(n)) == str(n)
    assert repr(wrapper(n)) == f"{wrapper.__name__}({n})"


@pytest.mark.crdb("skip", reason="all types returned as bigint? TODOCRDB")
@pytest.mark.parametrize("wrapper", "Int2 Int4 Int8 Oid Float4 Float8".split())
@pytest.mark.parametrize("fmt_in", PyFormat)
def test_repr_wrapper(conn, wrapper, fmt_in):
    wrapper = getattr(psycopg.types.numeric, wrapper)
    cur = conn.execute(f"select pg_typeof(%{fmt_in.value})::oid", [wrapper(0)])
    oid = cur.fetchone()[0]
    assert oid == psycopg.postgres.types[wrapper.__name__.lower()].oid


@pytest.mark.parametrize("fmt_out", pq.Format)
@pytest.mark.parametrize(
    "typename",
    "integer int2 int4 int8 float4 float8 numeric".split() + ["double precision"],
)
def test_oid_lookup(conn, typename, fmt_out):
    dumper = conn.adapters.get_dumper_by_oid(conn.adapters.types[typename].oid, fmt_out)
    assert dumper.oid == conn.adapters.types[typename].oid
    assert dumper.format == fmt_out



# WARNING: this file is auto-generated by 'async_to_sync.py'
# from the original file 'test_tpc_async.py'
# DO NOT CHANGE! Change the original file instead.
import pytest

import psycopg
from psycopg.pq import TransactionStatus

pytestmark = pytest.mark.crdb_skip("2-phase commit")


def test_tpc_disabled(conn, pipeline):
    cur = conn.execute("show max_prepared_transactions")

    if int(cur.fetchone()[0]):
        pytest.skip("prepared transactions enabled")

    conn.rollback()
    conn.tpc_begin("x")
    with pytest.raises(psycopg.NotSupportedError):
        conn.tpc_prepare()


class TestTPC:

    def test_tpc_commit(self, conn, tpc):
        xid = conn.xid(1, "gtrid", "bqual")
        assert conn.info.transaction_status == TransactionStatus.IDLE

        conn.tpc_begin(xid)
        assert conn.info.transaction_status == TransactionStatus.INTRANS

        cur = conn.cursor()
        cur.execute("insert into test_tpc values ('test_tpc_commit')")
        assert tpc.count_xacts() == 0
        assert tpc.count_test_records() == 0

        conn.tpc_prepare()
        assert conn.info.transaction_status == TransactionStatus.IDLE
        assert tpc.count_xacts() == 1
        assert tpc.count_test_records() == 0

        conn.tpc_commit()
        assert conn.info.transaction_status == TransactionStatus.IDLE
        assert tpc.count_xacts() == 0
        assert tpc.count_test_records() == 1

    def test_tpc_commit_one_phase(self, conn, tpc):
        xid = conn.xid(1, "gtrid", "bqual")
        assert conn.info.transaction_status == TransactionStatus.IDLE

        conn.tpc_begin(xid)
        assert conn.info.transaction_status == TransactionStatus.INTRANS

        cur = conn.cursor()
        cur.execute("insert into test_tpc values ('test_tpc_commit_1p')")
        assert tpc.count_xacts() == 0
        assert tpc.count_test_records() == 0

        conn.tpc_commit()
        assert conn.info.transaction_status == TransactionStatus.IDLE
        assert tpc.count_xacts() == 0
        assert tpc.count_test_records() == 1

    def test_tpc_commit_recovered(self, conn_cls, conn, dsn, tpc):
        xid = conn.xid(1, "gtrid", "bqual")
        assert conn.info.transaction_status == TransactionStatus.IDLE

        conn.tpc_begin(xid)
        assert conn.info.transaction_status == TransactionStatus.INTRANS

        cur = conn.cursor()
        cur.execute("insert into test_tpc values ('test_tpc_commit_rec')")
        assert tpc.count_xacts() == 0
        assert tpc.count_test_records() == 0

        conn.tpc_prepare()
        conn.close()
        assert tpc.count_xacts() == 1
        assert tpc.count_test_records() == 0

        with conn_cls.connect(dsn) as conn:
            xid = conn.xid(1, "gtrid", "bqual")
            conn.tpc_commit(xid)
            assert conn.info.transaction_status == TransactionStatus.IDLE

        assert tpc.count_xacts() == 0
        assert tpc.count_test_records() == 1

    def test_tpc_rollback(self, conn, tpc):
        xid = conn.xid(1, "gtrid", "bqual")
        assert conn.info.transaction_status == TransactionStatus.IDLE

        conn.tpc_begin(xid)
        assert conn.info.transaction_status == TransactionStatus.INTRANS

        cur = conn.cursor()
        cur.execute("insert into test_tpc values ('test_tpc_rollback')")
        assert tpc.count_xacts() == 0
        assert tpc.count_test_records() == 0

        conn.tpc_prepare()
        assert conn.info.transaction_status == TransactionStatus.IDLE
        assert tpc.count_xacts() == 1
        assert tpc.count_test_records() == 0

        conn.tpc_rollback()
        assert conn.info.transaction_status == TransactionStatus.IDLE
        assert tpc.count_xacts() == 0
        assert tpc.count_test_records() == 0

    def test_tpc_rollback_one_phase(self, conn, tpc):
        xid = conn.xid(1, "gtrid", "bqual")
        assert conn.info.transaction_status == TransactionStatus.IDLE

        conn.tpc_begin(xid)
        assert conn.info.transaction_status == TransactionStatus.INTRANS

        cur = conn.cursor()
        cur.execute("insert into test_tpc values ('test_tpc_rollback_1p')")
        assert tpc.count_xacts() == 0
        assert tpc.count_test_records() == 0

        conn.tpc_rollback()
        assert conn.info.transaction_status == TransactionStatus.IDLE
        assert tpc.count_xacts() == 0
        assert tpc.count_test_records() == 0

    def test_tpc_rollback_recovered(self, conn_cls, conn, dsn, tpc):
        xid = conn.xid(1, "gtrid", "bqual")
        assert conn.info.transaction_status == TransactionStatus.IDLE

        conn.tpc_begin(xid)
        assert conn.info.transaction_status == TransactionStatus.INTRANS

        cur = conn.cursor()
        cur.execute("insert into test_tpc values ('test_tpc_commit_rec')")
        assert tpc.count_xacts() == 0
        assert tpc.count_test_records() == 0

        conn.tpc_prepare()
        conn.close()
        assert tpc.count_xacts() == 1
        assert tpc.count_test_records() == 0

        with conn_cls.connect(dsn) as conn:
            xid = conn.xid(1, "gtrid", "bqual")
            conn.tpc_rollback(xid)
            assert conn.info.transaction_status == TransactionStatus.IDLE

        assert tpc.count_xacts() == 0
        assert tpc.count_test_records() == 0

    def test_status_after_recover(self, conn, tpc):
        assert conn.info.transaction_status == TransactionStatus.IDLE
        conn.tpc_recover()
        assert conn.info.transaction_status == TransactionStatus.IDLE

        cur = conn.cursor()
        cur.execute("select 1")
        assert conn.info.transaction_status == TransactionStatus.INTRANS
        conn.tpc_recover()
        assert conn.info.transaction_status == TransactionStatus.INTRANS

    def test_recovered_xids(self, conn, tpc):
        # insert a few test xns
        conn.set_autocommit(True)
        cur = conn.cursor()
        cur.execute("begin; prepare transaction '1-foo'")
        cur.execute("begin; prepare transaction '2-bar'")

        # read the values to return
        cur.execute(
            """
            select gid, prepared, owner, database from pg_prepared_xacts
            where database = %s
            """,
            (conn.info.dbname,),
        )
        okvals = cur.fetchall()
        okvals.sort()

        xids = conn.tpc_recover()
        xids = [xid for xid in xids if xid.database == conn.info.dbname]
        xids.sort(key=lambda x: x.gtrid)

        # check the values returned
        assert len(okvals) == len(xids)
        for xid, (gid, prepared, owner, database) in zip(xids, okvals):
            assert xid.gtrid == gid
            assert xid.prepared == prepared
            assert xid.owner == owner
            assert xid.database == database

    def test_xid_encoding(self, conn, tpc):
        xid = conn.xid(42, "gtrid", "bqual")
        conn.tpc_begin(xid)
        conn.tpc_prepare()

        cur = conn.cursor()
        cur.execute(
            "select gid from pg_prepared_xacts where database = %s", (conn.info.dbname,)
        )
        assert "42_Z3RyaWQ=_YnF1YWw=" == cur.fetchone()[0]

    @pytest.mark.parametrize(
        "fid, gtrid, bqual",
        [(0, "", ""), (42, "gtrid", "bqual"), (2147483647, "x" * 64, "y" * 64)],
    )
    def test_xid_roundtrip(self, conn_cls, conn, dsn, tpc, fid, gtrid, bqual):
        xid = conn.xid(fid, gtrid, bqual)
        conn.tpc_begin(xid)
        conn.tpc_prepare()
        conn.close()

        with conn_cls.connect(dsn) as conn:
            xids = [x for x in conn.tpc_recover() if x.database == conn.info.dbname]
            assert len(xids) == 1
            xid = xids[0]
            conn.tpc_rollback(xid)

        assert xid.format_id == fid
        assert xid.gtrid == gtrid
        assert xid.bqual == bqual

    # 199 is PostgreSQL's limit in transaction id length

    @pytest.mark.parametrize("tid", ["", "hello, world!", "x" * 199])
    def test_unparsed_roundtrip(self, conn_cls, conn, dsn, tpc, tid):
        conn.tpc_begin(tid)
        conn.tpc_prepare()
        conn.close()

        with conn_cls.connect(dsn) as conn:
            xids = [x for x in conn.tpc_recover() if x.database == conn.info.dbname]
            assert len(xids) == 1
            xid = xids[0]
            conn.tpc_rollback(xid)

        assert xid.format_id is None
        assert xid.gtrid == tid
        assert xid.bqual is None

    def test_xid_unicode(self, conn_cls, conn, dsn, tpc):
        x1 = conn.xid(10, "uni", "code")
        conn.tpc_begin(x1)
        conn.tpc_prepare()
        conn.close()

        with conn_cls.connect(dsn) as conn:
            xid = [x for x in conn.tpc_recover() if x.database == conn.info.dbname][0]

        assert 10 == xid.format_id
        assert "uni" == xid.gtrid
        assert "code" == xid.bqual

    def test_xid_unicode_unparsed(self, conn_cls, conn, dsn, tpc):
        # We don't expect people shooting snowmen as transaction ids,
        # so if something explodes in an encode error I don't mind.
        # Let's just check unicode is accepted as type.
        conn.execute("set client_encoding to utf8")
        conn.commit()

        conn.tpc_begin("transaction-id")
        conn.tpc_prepare()
        conn.close()

        with conn_cls.connect(dsn) as conn:
            xid = [x for x in conn.tpc_recover() if x.database == conn.info.dbname][0]

        assert xid.format_id is None
        assert xid.gtrid == "transaction-id"
        assert xid.bqual is None

    def test_cancel_fails_prepared(self, conn, tpc):
        conn.tpc_begin("cancel")
        conn.tpc_prepare()
        with pytest.raises(psycopg.ProgrammingError):
            conn.cancel()
        with pytest.raises(psycopg.ProgrammingError):
            conn.cancel_safe()

    def test_tpc_recover_non_dbapi_connection(self, conn_cls, conn, dsn, tpc):
        conn.row_factory = psycopg.rows.dict_row
        conn.tpc_begin("dict-connection")
        conn.tpc_prepare()
        conn.close()

        with conn_cls.connect(dsn) as conn:
            xids = conn.tpc_recover()
            xid = [x for x in xids if x.database == conn.info.dbname][0]

        assert xid.format_id is None
        assert xid.gtrid == "dict-connection"
        assert xid.bqual is None



"""
psycopg_c._psycopg optimization module.

The module contains optimized C code used in preference to Python code
if a compiler is available.
"""

# Copyright (C) 2020 The Psycopg Team

from psycopg_c cimport pq
from psycopg_c.pq cimport libpq
from psycopg_c._psycopg cimport oids

import logging

from psycopg.pq import Format as _pq_Format
from psycopg._enums import PyFormat as _py_Format

logger = logging.getLogger("psycopg")

PQ_TEXT = _pq_Format.TEXT
PQ_BINARY = _pq_Format.BINARY

PG_AUTO = _py_Format.AUTO
PG_TEXT = _py_Format.TEXT
PG_BINARY = _py_Format.BINARY


cdef extern from *:
    """
/* Include this early to avoid a warning about redefined ARRAYSIZE in winnt.h */
#ifdef MS_WINDOWS
#define WIN32_LEAN_AND_MEAN
#include <winsock2.h>
#endif

#ifndef ARRAYSIZE
#define ARRAYSIZE(a) ((sizeof(a) / sizeof(*(a))))
#endif
    """
    int ARRAYSIZE(void *array)


include "_psycopg/adapt.pyx"
include "_psycopg/copy.pyx"
include "_psycopg/generators.pyx"
include "_psycopg/transform.pyx"
include "_psycopg/waiting.pyx"

include "types/array.pyx"
include "types/datetime.pyx"
include "types/numeric.pyx"
include "types/bool.pyx"
include "types/numpy.pyx"
include "types/string.pyx"
include "types/uuid.pyx"



"""
psycopg_c.pq.Escaping object implementation.
"""

# Copyright (C) 2020 The Psycopg Team

from cpython.mem cimport PyMem_Free, PyMem_Malloc
from libc.string cimport strlen


cdef class Escaping:
    def __init__(self, PGconn conn = None):
        self.conn = conn

    cpdef escape_literal(self, data):
        cdef char *out
        cdef char *ptr
        cdef Py_ssize_t length

        if self.conn is None:
            raise e.OperationalError("escape_literal failed: no connection provided")
        if self.conn._pgconn_ptr is NULL:
            raise e.OperationalError("the connection is closed")

        _buffer_as_string_and_size(data, &ptr, &length)

        out = libpq.PQescapeLiteral(self.conn._pgconn_ptr, ptr, length)
        if out is NULL:
            raise e.OperationalError(
                f"escape_literal failed: {self.conn.get_error_message()}"
            )

        rv = out[:strlen(out)]
        libpq.PQfreemem(out)
        return rv

    cpdef escape_identifier(self, data):
        cdef char *out
        cdef char *ptr
        cdef Py_ssize_t length

        _buffer_as_string_and_size(data, &ptr, &length)

        if self.conn is None:
            raise e.OperationalError("escape_identifier failed: no connection provided")
        if self.conn._pgconn_ptr is NULL:
            raise e.OperationalError("the connection is closed")

        out = libpq.PQescapeIdentifier(self.conn._pgconn_ptr, ptr, length)
        if out is NULL:
            raise e.OperationalError(
                f"escape_identifier failed: {self.conn.get_error_message()}"
            )

        rv = out[:strlen(out)]
        libpq.PQfreemem(out)
        return rv

    cpdef escape_string(self, data):
        cdef int error
        cdef size_t len_out
        cdef char *ptr
        cdef char *buf_out
        cdef Py_ssize_t length

        _buffer_as_string_and_size(data, &ptr, &length)

        if self.conn is not None:
            if self.conn._pgconn_ptr is NULL:
                raise e.OperationalError("the connection is closed")

            buf_out = <char *>PyMem_Malloc(length * 2 + 1)
            len_out = libpq.PQescapeStringConn(
                self.conn._pgconn_ptr, buf_out, ptr, length, &error
            )
            if error:
                PyMem_Free(buf_out)
                raise e.OperationalError(
                    f"escape_string failed: {self.conn.get_error_message()}"
                )

        else:
            buf_out = <char *>PyMem_Malloc(length * 2 + 1)
            len_out = libpq.PQescapeString(buf_out, ptr, length)

        rv = buf_out[:len_out]
        PyMem_Free(buf_out)
        return rv

    cpdef escape_bytea(self, data):
        cdef size_t len_out
        cdef unsigned char *out
        cdef char *ptr
        cdef Py_ssize_t length

        if self.conn is not None and self.conn._pgconn_ptr is NULL:
            raise e.OperationalError("the connection is closed")

        _buffer_as_string_and_size(data, &ptr, &length)

        if self.conn is not None:
            out = libpq.PQescapeByteaConn(
                self.conn._pgconn_ptr, <unsigned char *>ptr, length, &len_out)
        else:
            out = libpq.PQescapeBytea(<unsigned char *>ptr, length, &len_out)

        if out is NULL:
            raise MemoryError(
                f"couldn't allocate for escape_bytea of {len(data)} bytes"
            )

        rv = out[:len_out - 1]  # out includes final 0
        libpq.PQfreemem(out)
        return rv

    cpdef unescape_bytea(self, const unsigned char *data):
        # not needed, but let's keep it symmetric with the escaping:
        # if a connection is passed in, it must be valid.
        if self.conn is not None:
            if self.conn._pgconn_ptr is NULL:
                raise e.OperationalError("the connection is closed")

        cdef size_t len_out
        cdef unsigned char *out = libpq.PQunescapeBytea(data, &len_out)
        if out is NULL:
            raise MemoryError(
                f"couldn't allocate for unescape_bytea of {len(data)} bytes"
            )

        rv = out[:len_out]
        libpq.PQfreemem(out)
        return rv



#!/usr/bin/env python
"""Wrapper for pproxy to fix Python 3.14 compatibility

Work around https://github.com/qwj/python-proxy/issues/200
"""

import sys
import asyncio
from typing import Any

from pproxy.server import main as main_  # type: ignore[import-untyped]


def main() -> Any:
    # Before Python 3.14 `get_event_loop()` used to create a new loop.
    # From Python 3.14 it raises a `RuntimeError`.
    try:
        asyncio.get_event_loop()
    except RuntimeError:
        asyncio.set_event_loop(asyncio.new_event_loop())

    return main_()


if __name__ == "__main__":
    sys.exit(main())
